"""
Assists in grabbing the requisite data
"""

import copy
import glob
import json
import os

from ..models.common_models import Molecule

__all__ = ["list_directories", "get_file_name", "get_file", "get_molecule", "get_options"]

_data_dir = os.path.dirname(__file__)

_folders = ["molecules", "options"]
_data_folders = {x: os.path.join(_data_dir, x) for x in _folders}


def _get_folder_path(folder):
    if folder not in _data_folders:
        raise KeyError("Folder '{}' not recognized".format(folder))

    return _data_folders[folder]


def list_directories():
    """
    List all known directories.
    """
    return copy.deepcopy(_data_folders.keys())


def get_file_name(folder, filename=None):
    folder = _get_folder_path(folder)
    if filename:
        folder = os.path.join(folder, filename)

    files = glob.glob(folder)
    if len(files) == 1:
        return files[0]
    else:
        return files


def get_file(folder, *args):
    folder = _get_folder_path(folder)
    filename = os.path.join(folder, *args)
    if not os.path.isfile(filename):
        raise OSError("Path '{}' not found.".format(filename))

    with open(filename, "r") as infile:
        ret = infile.read()

    return ret


def get_molecule(name, orient=True):
    """
    Returns a Molecule object from the available stored objects.
    """
    fname = get_file_name("molecules", name)
    if not fname:
        raise OSError("File: {}/{} not found".format("molecules", name))

    return Molecule.from_file(fname, orient=orient)


def get_options(name):
    """
    Returns a default options dictionary
    """
    folder = _get_folder_path("options")
    if ".json" not in name:
        name += ".json"

    with open(os.path.join(folder, name), "r") as infile:
        ret = json.load(infile)
    return ret



"""
Web handlers for the FractalServer
"""
import json

import tornado.web

from .interface.models.rest_models import (
    MoleculeGETBody, MoleculeGETResponse, MoleculePOSTBody, MoleculePOSTResponse,
    KeywordGETBody, KeywordGETResponse, KeywordPOSTBody, KeywordPOSTResponse,
    CollectionGETBody, CollectionGETResponse, CollectionPOSTBody, CollectionPOSTResponse,
    ResultGETBody, ResultGETResponse,
    ProcedureGETBody, ProcedureGETReponse
)


class APIHandler(tornado.web.RequestHandler):
    """
    A requests handler for API calls, build
    """

    def initialize(self, **objects):
        """
        Initializes the request to JSON, adds objects, and logging.
        """

        self.set_header("Content-Type", "application/json")
        self.objects = objects
        self.logger = objects["logger"]

        #print(self.request.headers["Content-Type"])
        self.json = json.loads(self.request.body.decode("UTF-8"))

    def authenticate(self, permission):
        """Authenticates request with a given permission setting

        Parameters
        ----------
        permission : str
            The required permission ["read", "write", "compute", "admin"]

        """
        if "Authorization" in self.request.headers:

            data = json.loads(self.request.headers["Authorization"])
            username = data["username"]
            password = data["password"]
        else:
            username = None
            password = None

        verified, msg = self.objects["storage_socket"].verify_user(username, password, permission)
        if verified is False:
            raise tornado.web.HTTPError(status_code=401, reason=msg)


class InformationHandler(APIHandler):
    """
    A handler that returns public server information
    """

    def get(self):
        """

        """
        self.authenticate("read")

        self.logger.info("GET: Information")

        self.write(self.objects["public_information"])


class MoleculeHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    def get(self):
        """

        Experimental documentation, need to find a decent format.

        Request:
            "meta" - Overall options to the Molecule pull request
                - "index" - What kind of index used to find the data ("id", "molecule_hash", "molecular_formula")
            "data" - A dictionary of {key : index} requests

        Returns:
            "meta" - Metadata associated with the query
                - "errors" - A list of errors in (index, error_id) format.
                - "n_found" - The number of molecule found.
                - "success" - If the query was successful or not.
                - "error_description" - A string based description of the error or False
                - "missing" - A list of keys that were not found.
            "data" - A dictionary of {key : molecule JSON} results

        """
        self.authenticate("read")

        storage = self.objects["storage_socket"]

        body = MoleculeGETBody.parse_raw(self.request.body)

        molecules = storage.get_molecules(body.data, index=body.meta.index.value)
        self.logger.info("GET: Molecule - {} pulls.".format(len(molecules["data"])))

        ret = MoleculeGETResponse(**molecules)
        self.write(ret.json())

    def post(self):
        """
            Experimental documentation, need to find a decent format.

        Request:
            "meta" - Overall options to the Molecule pull request
                - No current options
            "data" - A dictionary of {key : molecule JSON} requests

        Returns:
            "meta" - Metadata associated with the query
                - "errors" - A list of errors in (index, error_id) format.
                - "n_inserted" - The number of molecule inserted.
                - "success" - If the query was successful or not.
                - "error_description" - A string based description of the error or False
                - "duplicates" - A list of keys that were already inserted.
            "data" - A dictionary of {key : id} results
        """

        self.authenticate("write")

        storage = self.objects["storage_socket"]

        body = MoleculePOSTBody.parse_raw(self.request.body)
        ret = storage.add_molecules(body.data)
        response = MoleculePOSTResponse(**ret)

        self.logger.info("POST: Molecule - {} inserted.".format(response.meta.n_inserted))

        self.write(response.json())


class OptionHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    def get(self):
        self.authenticate("read")

        storage = self.objects["storage_socket"]

        body = KeywordGETBody.parse_raw(self.request.body)
        ret = storage.get_keywords(**body.data, with_ids=False)
        kw = KeywordGETResponse(**ret)

        self.logger.info("GET: Keywords - {} pulls.".format(len(kw.data)))

        self.write(kw.json())

    def post(self):
        self.authenticate("write")

        storage = self.objects["storage_socket"]

        body = KeywordPOSTBody.parse_raw(self.request.body)
        ret = storage.add_keywords([x.json_dict() for x in body.data])
        response = KeywordPOSTResponse(**ret)

        self.logger.info("POST: Keywords - {} inserted.".format(response.meta.n_inserted))

        self.write(response.json())


class CollectionHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    def get(self):
        self.authenticate("read")

        storage = self.objects["storage_socket"]

        body = CollectionGETBody.parse_raw(self.request.body)
        cols = storage.get_collections(**body.data.dict())
        response = CollectionGETResponse(**cols)
        self.logger.info("GET: Keywords - {} pulls.".format(len(response.data)))

        self.write(response.json())

    def post(self):
        self.authenticate("write")

        storage = self.objects["storage_socket"]

        body = CollectionPOSTBody.parse_raw(self.request.body)
        ret = storage.add_collection(body.data.collection,
                                     body.data.name,
                                     body.data.dict(exclude={"collection", "name"}),
                                     overwrite=body.meta.overwrite)

        response = CollectionPOSTResponse(**ret)

        self.logger.info("POST: Collections - {} inserted.".format(response.meta.n_inserted))

        self.write(response.json())


class ResultHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    def get(self):
        self.authenticate("read")

        storage = self.objects["storage_socket"]

        body = ResultGETBody.parse_raw(self.request.body)
        proj = body.meta.projection
        if 'id' in body.data:
            ret = storage.get_results_by_id(body.data['id'], projection=proj)
        elif 'task_id' in body.data:
            ret = storage.get_results_by_task_id(body.data['task_id'], projection=proj)
        else:
            ret = storage.get_results(**body.data, projection=proj)
        result = ResultGETResponse(**ret)
        self.logger.info("GET: Results - {} pulls.".format(len(result.data)))
        self.write(result.json())


class ProcedureHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    def get(self):
        self.authenticate("read")

        storage = self.objects["storage_socket"]

        body = ProcedureGETBody.parse_raw(self.request.body)

        if "id" in body.data:
            ret = storage.get_procedures_by_id(id=body.data["id"])
        elif "hash_index" in body.data:
            ret = storage.get_procedures_by_id(hash_index=body.data["hash_index"])
        elif 'task_id' in body.data:
            ret = storage.get_procedures_by_task_id(body.data["task_id"])
        else:
            ret = storage.get_procedures(**body.data)

        response = ProcedureGETReponse(**ret)
        self.logger.info("GET: Procedures - {} pulls.".format(len(response.data)))

        self.write(response.json())




"""A module for statistical quantities.
"""
import numpy as np
import pandas as pd

# Need to use specifically with the fitting_database class

# def signed_error(value, bench):
#     return value - bench

# def unsigned_error(value, bench):
#     return np.abs(value - bench)


def mean_signed_error(value, bench):
    return np.mean(value - bench)


def mean_unsigned_error(value, bench):
    return np.mean(np.abs(value - bench))


# def unsigned_relative_error(value, bench):
#     return np.abs((value - bench) / bench) * 100


def mean_unsigned_relative_error(value, bench):
    return np.mean(np.abs((value - bench) / bench)) * 100


# def weighted_unsigned_relative_error(value, bench, weight):
#     return np.abs((value - bench) / weight) * 100

# def weighted_mean_unsigned_relative_error(value, bench, weight):
#     return np.mean(np.abs((value - bench) / weight)) * 100

# Stats wrapped in a dictionary:
_stats_dict = {}
# _stats_dict['E'] = signed_error
_stats_dict['ME'] = mean_signed_error
# _stats_dict['UE'] = unsigned_error
_stats_dict['MUE'] = mean_unsigned_error
# _stats_dict['URE'] = unsigned_relative_error
_stats_dict['MURE'] = mean_unsigned_relative_error
# _stats_dict['WURE'] = weighted_unsigned_relative_error
# _stats_dict['WMURE'] = weighted_mean_unsigned_relative_error

_return_series = ['ME', 'MUE', 'MURE', 'WMURE']

# _needs_weight = ["WURE", "WMURE"]


def wrap_statistics(description, df, value, bench):

    # Get benchmark
    if isinstance(bench, str):
        rbench = df[bench]
    elif isinstance(bench, (np.ndarray, pd.Series)):
        if len(bench.shape) != 1:
            raise ValueError('Only 1D numpy arrays can be passed to statistical quantities.')
        rbench = bench
    else:
        raise TypeError('Benchmark must a column of the dataframe or a 1D numpy array.')

    if isinstance(value, str):
        rvalue = df[value]
        return _stats_dict[description](rvalue, rbench)

    elif isinstance(value, pd.Series):
        rvalue = value
        return _stats_dict[description](rvalue, rbench)

    elif isinstance(value, pd.DataFrame):
        return value.apply(lambda x: _stats_dict[description](x, rbench))

    elif isinstance(value, (list, tuple)):
        if description in _return_series:
            ret = pd.Series(index=value)
        else:
            ret = pd.DataFrame(columns=value)

        method = _stats_dict[description]
        for col in value:
            ret[col] = method(df[col], rbench)
        return ret

    else:
        raise TypeError('Type {} is not understood for statistical quantities'.format(str(type(value))))



"""
    Unit tests of the mongoengine interface to MongoDB

    Does not use portal or fractal server interfaces.

"""

import qcfractal.interface as portal

from qcfractal.storage_sockets.models import Molecule, Result, Keywords, \
                    Procedure, OptimizationProcedure, TorsiondriveProcedure
from qcfractal.storage_sockets.models import TaskQueue
from qcfractal.testing import mongoengine_socket_fixture as storage_socket
import pytest
from bson import ObjectId


@pytest.fixture
def molecules_H4O2(storage_socket):
    water = portal.data.get_molecule("water_dimer_minima.psimol")
    water2 = portal.data.get_molecule("water_dimer_stretch.psimol")

    ret = storage_socket.add_molecules({"water1": water.json_dict(),
                                        "water2": water2.json_dict()})

    yield list(ret['data'].values())

    storage_socket.del_molecules(water.get_hash(), index="hash")
    storage_socket.del_molecules(water2.get_hash(), index="hash")


def test_molecule(storage_socket):
    """
        Test the use of the ME class Molecule

        Note:
            creation of a Molecule using ME is not implemented yet
            Should create a Molecule using: mongoengine_socket.add_molecules
    """

    # don't use len(Molecule.objects), slow
    num_mol_in_db = Molecule.objects().count()
    # Molecule.objects().delete()
    assert num_mol_in_db == 0

    water = portal.data.get_molecule("water_dimer_minima.psimol")
    water2 = portal.data.get_molecule("water_dimer_stretch.psimol")

    # Add Molecule using pymongo
    ret = storage_socket.add_molecules({"water1": water.json_dict(),
                                            "water2": water2.json_dict()})
    assert ret["meta"]["success"] is True
    assert ret["meta"]["n_inserted"] == 2

    # Use the ORM class
    water_mol = Molecule.objects().first()
    assert water_mol.molecular_formula == "H4O2"
    assert water_mol.molecular_charge == 0

    # print(water_mol.json_dict())

    # Query with fields in the model
    result_list = Molecule.objects(molecular_formula="H4O2")
    assert len(result_list) == 2
    assert result_list[0].molecular_multiplicity == 1

    # Query with fields NOT in the model. works too!
    result_list = Molecule.objects(molecular_charge=0)
    assert len(result_list) == 2

    # get unique by hash and formula
    one_mol = Molecule.objects(molecule_hash=water_mol.molecule_hash,
                               molecular_formula=water_mol.molecular_formula)
    assert len(one_mol) == 1

    # Clean up
    storage_socket.del_molecules(water.get_hash(), index="hash")
    storage_socket.del_molecules(water2.get_hash(), index="hash")


def test_results(storage_socket, molecules_H4O2):
    """
        Handling results throught the ME classes
    """

    assert Result.objects().count() == 0
    assert Keywords.objects().count() == 0

    molecules = molecules_H4O2

    assert len(molecules) == 2

    page1 = {
        "molecule": ObjectId(molecules[0]),
        "method": "M1",
        "basis": "B1",
        "keywords": None,
        "program": "P1",
        "driver": "energy",
        "other_data": 5,
    }

    page2 = {
        "molecule": ObjectId(molecules[1]),
        "method": "M1",
        "basis": "B1",
        "keywords": None,
        "program": "P1",
        "driver": "energy",
        "other_data": 10,
    }

    Result(**page1).save()
    ret = Result.objects(method='M1').first()
    assert ret.molecule.molecular_formula == 'H4O2'

    # clean up
    Result.objects().delete()


def test_procedure(storage_socket):
    """
        Handling procedure throught the ME classes
    """

    assert Procedure.objects().count() == 0
    # assert Keywords.objects().count() == 0

    # molecules = Molecule.objects(molecular_formula='H4O2')
    # assert molecules.count() == 2

    data1 = {
        # "molecule": molecules[0],
        "procedure": "undefined",
        "keywords": None,
        "program": "P5",
        "qc_meta": {
            "basis": "B1",
            "program": "P1",
            "method": "M1",
            "driver": "energy"
        },
    }

    procedure = Procedure(**data1)
    procedure.save()
    assert procedure.id
    # print('Procedure After save: ', procedure.json_dict())
    # assert procedure.molecule.molecular_formula == 'H4O2'


def test_optimization_procedure(storage_socket, molecules_H4O2):
    """
        Optimization procedure
    """

    assert OptimizationProcedure.objects().count() == 0
    # assert Keywords.objects().count() == 0

    data1 = {
        "initial_molecule": ObjectId(molecules_H4O2[0]),
        # "procedure_type": None,
        "keywords": None,
        "program": "P7",
        "qc_meta": {
            "basis": "B1",
            "program": "P1",
            "method": "M1",
            "driver": "energy"
        },
    }

    procedure = OptimizationProcedure(**data1).save()
    proc = OptimizationProcedure.objects().first()
    assert proc.initial_molecule.molecular_formula == 'H4O2'


def test_torsiondrive_procedure(storage_socket):
    """
        Torsiondrive procedure
    """

    assert TorsiondriveProcedure.objects().count() == 0
    # assert Keywords.objects().count() == 0

    # molecules = Molecule.objects(molecular_formula='H4O2')
    # assert molecules.count() == 2

    data1 = {
        # "molecule": molecules[0],
        # "procedure": None,
        "keywords": None,
        "program": "P9",
        "qc_meta": {
            "basis": "B1",
            "program": "P1",
            "method": "M1",
            "driver": "energy"
        },
    }

    procedure = TorsiondriveProcedure(**data1)
    procedure.save()
    # print('TorsiondriveProcedure After save: ', procedure.json_dict())


def test_add_task_queue(storage_socket, molecules_H4O2):
    """
        Simple test of adding a task using the ME classes
        in QCFractal, tasks should be added using storage_socket
    """

    assert TaskQueue.objects.count() == 0
    TaskQueue.objects().delete()

    page1 = {
        "molecule": ObjectId(molecules_H4O2[0]),
        "method": "M1",
        "basis": "B1",
        "keywords": None,
        "program": "P1",
        "driver": "energy",
        "other_data": 5,
    }
    # add a task that reference results
    result = Result(**page1).save()

    task = TaskQueue(base_result=result)
    task.save()
    assert TaskQueue.objects().count() == 1

    # add a task that reference Optimization Procedure
    opt = OptimizationProcedure.objects().first()

    task = TaskQueue(base_result=opt)
    task.save()
    assert TaskQueue.objects().count() == 2

    # add a task that reference Torsiondrive Procedure
    data1 = {
        "keywords": None,
        "program": "P9",
        "qc_meta": {
            "basis": "B1",
            "program": "P1",
            "method": "M1",
            "driver": "energy"
        },
    }

    tor = TorsiondriveProcedure(**data1).save()

    task = TaskQueue(base_result=tor)
    task.save()
    assert TaskQueue.objects().count() == 3


def test_queue(storage_socket):
    tasks = TaskQueue.objects(status='WAITING')\
                .limit(1000)\
                .order_by('-created_on')\
                .select_related()   # *** no lazy load of ReferenceField, get them now (trurns of dereferencing, max_depth=1)
                # .only(projections_list)
                # .fields(..)
                # .exculde(..)
                # .no_dereference()  # don't get any of the ReferenceFields (ids) (Turning off dereferencing)
    assert len(tasks) == 3



"""
Tests the database wrappers

All tests should be atomic, that is create and cleanup their data
"""

import pytest

import qcfractal.interface as portal
from qcfractal.testing import mongoengine_socket_fixture as storage_socket


def test_molecules_add(storage_socket):

    water = portal.data.get_molecule("water_dimer_minima.psimol")

    # Add once
    ret1 = storage_socket.add_molecules({"new_water": water.json_dict()})
    assert ret1["meta"]["success"] is True
    assert ret1["meta"]["n_inserted"] == 1

    # Try duplicate adds
    ret2 = storage_socket.add_molecules({"new_water2": water.json_dict()})
    assert ret2["meta"]["success"] is True
    assert ret2["meta"]["n_inserted"] == 0
    assert ret2["meta"]["duplicates"][0] == "new_water2"

    # Assert the ids match
    assert ret1["data"]["new_water"] == ret2["data"]["new_water2"]

    # Pull molecule from the DB for tests
    db_json = storage_socket.get_molecules(water.get_hash(), index="hash")["data"][0]
    water.compare(db_json)

    # Cleanup adds
    ret = storage_socket.del_molecules(water.get_hash(), index="hash")
    assert ret == 1


def test_identical_mol_insert(storage_socket):
    """
    Tests as edge case where to identical molecules are added under different tags.
    """

    water = portal.data.get_molecule("water_dimer_minima.psimol")

    # Add two idential molecules
    ret1 = storage_socket.add_molecules({"w1": water.json_dict(), "w2": water.json_dict()})
    assert ret1["meta"]["success"] is True
    assert ret1["meta"]["n_inserted"] == 1
    assert ret1["data"]["w1"] == ret1["data"]["w2"]

    # Should only find one molecule
    ret2 = storage_socket.get_molecules([water.get_hash()], index="hash")
    assert ret2["meta"]["n_found"] == 1

    ret = storage_socket.del_molecules(water.get_hash(), index="hash")
    assert ret == 1


def test_molecules_add_many(storage_socket):
    water = portal.data.get_molecule("water_dimer_minima.psimol")
    water2 = portal.data.get_molecule("water_dimer_stretch.psimol")

    ret = storage_socket.add_molecules({"water1": water.json_dict(), "water2": water2.json_dict()})
    assert ret["meta"]["n_inserted"] == 2

    # Cleanup adds
    ret = storage_socket.del_molecules([water.get_hash(), water2.get_hash()], index="hash")
    assert ret == 2

    ret = storage_socket.add_molecules({"water1": water.json_dict(), "water2": water2.json_dict()})
    assert ret["meta"]["n_inserted"] == 2

    # Cleanup adds
    ret = storage_socket.del_molecules(list(ret["data"].values()), index="id")
    assert ret == 2


def test_molecules_get(storage_socket):

    water = portal.data.get_molecule("water_dimer_minima.psimol")

    # Add once
    ret = storage_socket.add_molecules({"water": water.json_dict()})
    assert ret["meta"]["n_inserted"] == 1
    water_id = ret["data"]["water"]

    # Pull molecule from the DB for tests
    db_json = storage_socket.get_molecules(water_id, index="id")["data"][0]
    water2 = portal.Molecule(**db_json)
    water2.compare(water)

    # Cleanup adds
    ret = storage_socket.del_molecules(water_id, index="id")
    assert ret == 1


def test_molecules_mixed_add_get(storage_socket):
    water = portal.data.get_molecule("water_dimer_minima.psimol")

    ret = storage_socket.get_add_molecules_mixed(["bad_id", water, "bad_id2"])
    assert ret["data"][0] is None
    assert ret["data"][1]["identifiers"]["molecule_hash"] == water.get_hash()
    assert ret["data"][2] is None
    assert set(ret["meta"]["missing"]) == {0, 2}

    # Cleanup adds
    ret = storage_socket.del_molecules([ret["data"][1]["id"]], index="id")
    assert ret == 1


def test_molecules_bad_get(storage_socket):

    water = portal.data.get_molecule("water_dimer_minima.psimol")

    # Add once
    ret = storage_socket.add_molecules({"water": water.json_dict()})
    assert ret["meta"]["n_inserted"] == 1
    water_id = ret["data"]["water"]

    # Pull molecule from the DB for tests
    ret = storage_socket.get_molecules([water_id, "something", 5, (3, 2)], index="id")
    assert len(ret["meta"]["errors"]) == 1
    assert ret["meta"]["errors"][0][0] == "Bad Ids"
    assert len(ret["meta"]["errors"][0][1]) == 3
    assert ret["meta"]["n_found"] == 1

    # Cleanup adds
    ret = storage_socket.del_molecules(water_id, index="id")
    assert ret == 1


def test_keywords_add(storage_socket):

    opts = {"program": "hello", "values": {"o": 5}, "hash_index": "something_unique"}

    ret = storage_socket.add_keywords([opts, opts.copy()])
    assert len(ret["data"]) == 2
    assert ret["meta"]["n_inserted"] == 1
    assert ret["data"][0] == ret["data"][1]

    ret = storage_socket.add_keywords(opts)
    assert ret["meta"]["n_inserted"] == 0

    ret = storage_socket.get_keywords(hash_index="something_unique")
    opts["id"] = ret["data"][0]["id"]
    assert ret["meta"]["n_found"] == 1
    assert ret["data"][0] == opts

    assert 1 == storage_socket.del_keywords(id=opts["id"])


def test_keywords_mixed_add_get(storage_socket):

    opts1 = portal.models.KeywordSet(**{"program": "hello", "values": {"o": 5}})
    id1 = storage_socket.add_keywords([opts1.json_dict()])["data"][0]

    opts2 = {"program": "hello", "values": {"o": 6}}
    opts = storage_socket.get_add_keywords_mixed([opts1, opts2, id1, "bad_id"])["data"]
    assert opts[0]["id"] == id1
    assert opts[1]["values"]["o"] == 6
    assert "id" in opts[1]
    assert opts[2]["id"] == id1
    assert opts[3] is None

    assert 1 == storage_socket.del_keywords(id=id1)
    assert 1 == storage_socket.del_keywords(id=opts[1]["id"])


def test_keywords_error(storage_socket):
    opts = {"program": "hello"}

    ret = storage_socket.add_keywords(opts)
    assert ret["meta"]["n_inserted"] == 0
    assert len(ret["meta"]["validation_errors"]) == 1


def test_collections_add(storage_socket):

    collection = 'TorsionDrive'
    name = 'Torsion123'
    db = {"something": "else", "array": ["54321"]}

    ret = storage_socket.add_collection(collection, name, db)

    assert ret["meta"]["n_inserted"] == 1

    ret = storage_socket.get_collections(collection, name)

    assert ret["meta"]["success"] == True
    assert ret["meta"]["n_found"] == 1
    assert db['something'] == ret["data"][0]['something']

    ret = storage_socket.del_collection(collection, name)
    assert ret == 1

    ret = storage_socket.get_collections(collection, "bleh")
    # assert len(ret["meta"]["missing"]) == 1
    assert ret["meta"]["n_found"] == 0


def test_collections_overwrite(storage_socket):

    collection = "TorsionDrive"
    name = "Torsion123"
    db = {"something": "else", "array": ["54321"]}

    ret = storage_socket.add_collection(collection, name, db)

    assert ret["meta"]["n_inserted"] == 1

    ret = storage_socket.get_collections(collection, name)
    assert ret["meta"]["n_found"] == 1

    db_update = {
        # "id": ret["data"][0]["id"],
        "collection": "TorsionDrive",  # no need to include
        "name": "Torsion123",  # no need to include
        "something": "New",
        "something2": "else",
        "array2": ["54321"]
    }
    ret = storage_socket.add_collection(collection, name, db_update, overwrite=True)
    assert ret["meta"]["success"] == True

    ret = storage_socket.get_collections(collection, name)
    assert ret["meta"]["n_found"] == 1

    # Check to make sure the field were replaced and not updated
    db_result = ret["data"][0]
    # existing fields will not be removed, the collection will be updated
    # You will need to remove the old collection and create a new one
    # assert "something" not in db_result
    assert "something" in db_result
    assert "something2" in db_result
    assert db_update['something'] == db_result['something']

    ret = storage_socket.del_collection(collection, name)
    assert ret == 1


def test_results_add(storage_socket):

    # Add two waters
    water = portal.data.get_molecule("water_dimer_minima.psimol")
    water2 = portal.data.get_molecule("water_dimer_stretch.psimol")
    mol_insert = storage_socket.add_molecules({"water1": water.json_dict(), "water2": water2.json_dict()})

    page1 = {
        "molecule": mol_insert["data"]["water1"],
        "method": "M1",
        "basis": "B1",
        "keywords": "default",
        "program": "P1",
        "driver": "energy",
        "other_data": 5,
        "hash_index": 0,
    }

    page2 = {
        "molecule": mol_insert["data"]["water2"],
        "method": "M1",
        "basis": "B1",
        "keywords": "default",
        "program": "P1",
        "driver": "energy",
        "other_data": 10,
        "hash_index": 1,
    }

    page3 = {
        "molecule": mol_insert["data"]["water2"],
        "method": "M22",
        "basis": "B1",
        "keywords": "default",
        "program": "P1",
        "driver": "energy",
        "other_data": 10,
        "hash_index": 2,
    }
    ids = []
    ret = storage_socket.add_results([page1, page2])
    assert ret["meta"]["n_inserted"] == 2
    ids.extend(ret['data'])

    # add with duplicates:
    ret = storage_socket.add_results([page1, page2, page3])

    assert ret["meta"]["n_inserted"] == 1
    assert len(ret['data']) == 3  # first 2 found are None
    assert len(ret["meta"]['duplicates']) == 2

    for res_id in ret['data']:
        if res_id is not None:
            ids.append(res_id)

    ret = storage_socket.del_results(ids)
    assert ret == 3
    ret = storage_socket.del_molecules(list(mol_insert["data"].values()), index="id")
    assert ret == 2


### Build out a set of query tests


@pytest.fixture(scope="module")
def storage_results(storage_socket):
    # Add two waters
    water = portal.data.get_molecule("water_dimer_minima.psimol")
    water2 = portal.data.get_molecule("water_dimer_stretch.psimol")
    mol_insert = storage_socket.add_molecules({"water1": water.json_dict(), "water2": water2.json_dict()})

    page1 = {
        "molecule": mol_insert["data"]["water1"],
        "method": "M1",
        "basis": "B1",
        "keywords": "default",
        "program": "P1",
        "driver": "energy",
        "return_result": 5,
        "hash_index": 0,
        "status": 'COMPLETE'
    }

    page2 = {
        "molecule": mol_insert["data"]["water2"],
        "method": "M1",
        "basis": "B1",
        "keywords": "default",
        "program": "P1",
        "driver": "energy",
        "return_result": 10,
        "hash_index": 1,
        "status": 'COMPLETE'
    }

    page3 = {
        "molecule": mol_insert["data"]["water1"],
        "method": "M1",
        "basis": "B1",
        "keywords": "default",
        "program": "P2",
        "driver": "gradient",
        "return_result": 15,
        "hash_index": 2,
        "status": 'COMPLETE'
    }

    page4 = {
        "molecule": mol_insert["data"]["water1"],
        "method": "M2",
        "basis": "B1",
        "keywords": "default",
        "program": "P2",
        "driver": "gradient",
        "return_result": 15,
        "hash_index": 3,
        "status": 'COMPLETE'
    }

    page5 = {
        "molecule": mol_insert["data"]["water2"],
        "method": "M2",
        "basis": "B1",
        "keywords": "default",
        "program": "P1",
        "driver": "gradient",
        "return_result": 20,
        "hash_index": 4,
        "status": 'COMPLETE'
    }

    page6 = {
        "molecule": mol_insert["data"]["water2"],
        "method": "M3",
        "basis": "B1",
        "keywords": "default",
        "program": "P1",
        "driver": "gradient",
        "return_result": 20,
        "hash_index": 5,
        "status": 'COMPLETE'
    }

    results_insert = storage_socket.add_results([page1, page2, page3, page4, page5, page6])
    assert results_insert["meta"]["n_inserted"] == 6

    yield storage_socket

    # Cleanup
    result_ids = [x for x in results_insert["data"]]
    ret = storage_socket.del_results(result_ids)
    assert ret == results_insert["meta"]["n_inserted"]

    ret = storage_socket.del_molecules(list(mol_insert["data"].values()), index="id")
    assert ret == mol_insert["meta"]["n_inserted"]


def test_results_query_total(storage_results):

    assert 6 == len(storage_results.get_results()["data"])


def test_get_results_by_ids(storage_results):
    results = storage_results.get_results()["data"]
    ids = [x['id'] for x in results]

    ret = storage_results.get_results_by_id(ids, return_json=False)
    assert ret["meta"]["n_found"] == 6
    assert len(ret["data"]) == 6

    ret = storage_results.get_results_by_id(ids, projection=['status'])
    assert ret['data'][0].keys() == {'id', 'status'}


def test_results_query_method(storage_results):

    ret = storage_results.get_results(method=["M2", "M1"])
    assert ret["meta"]["n_found"] == 5

    ret = storage_results.get_results(method=["M2"])
    assert ret["meta"]["n_found"] == 2

    ret = storage_results.get_results(method="M2")
    assert ret["meta"]["n_found"] == 2


def test_results_query_dual(storage_results):

    ret = storage_results.get_results(method=["M2", "M1"], program=["P1", "P2"])
    assert ret["meta"]["n_found"] == 5

    ret = storage_results.get_results(method=["M2"], program="P2")
    assert ret["meta"]["n_found"] == 1

    ret = storage_results.get_results(method="M2", program="P2")
    assert ret["meta"]["n_found"] == 1


def test_results_query_project(storage_results):
    """See new changes in design here"""

    ret = storage_results.get_results(method="M2", program="P2", projection={"return_result"})["data"][0]
    assert set(ret.keys()) == {"id", "return_result"}
    assert ret["return_result"] == 15

    # Note: explicitly set with_ids=False to remove ids
    ret = storage_results.get_results(
        method="M2", program="P2", with_ids=False, projection={"return_result"})["data"][0]
    assert set(ret.keys()) == {"return_result"}


def test_results_query_driver(storage_results):
    ret = storage_results.get_results(driver="energy")
    assert ret["meta"]["n_found"] == 2


# ------ New Task Queue tests ------
# No hash index, tasks are unique by their base_result


def test_queue_submit(storage_results):

    result1 = storage_results.get_results()['data'][0]

    task1 = {
        # "hash_index": idx,  # not used anymore
        "spec": {
            "function": "qcengine.compute_procedure",
            "args": [{
                "json_blob": "data"
            }],
            "kwargs": {},
        },
        "hooks": [("service", "x")],
        "tag": None,
        "base_result": ('results', result1['id'])
    }

    # Submit a new task
    ret = storage_results.queue_submit([task1])
    assert len(ret["data"]) == 1
    assert ret['meta']['n_inserted'] == 1

    # submit a duplicate task with a hook
    task1['hooks'] = [('service', 'y')]
    ret = storage_results.queue_submit([task1])
    assert len(ret["data"]) == 1
    assert ret['meta']['n_inserted'] == 0
    assert len(ret["meta"]['duplicates']) == 1


# ----------------------------------------------------------

# Builds tests for the queue - Changed design


def test_storage_queue_roundtrip(storage_results):

    result1 = storage_results.get_results()['data'][1]
    task1 = {
        # "hash_index": idx,
        "spec": {
            "function": "qcengine.compute_procedure",
            "args": [{
                "json_blob": "data"
            }],
            "kwargs": {},
        },
        "hooks": [("service", "")],
        "tag": None,
        "base_result": ('results', result1['id'])
    }

    # Submit a task
    r = storage_results.queue_submit([task1])
    assert len(r["data"]) == 1

    # Query for next tasks
    r = storage_results.queue_get_next("test_manager")
    assert r[0]["spec"]["function"] == task1["spec"]["function"]
    queue_id = r[0]["id"]

    # Mark task as done
    r = storage_results.queue_mark_complete([queue_id])
    assert r == 1

    # Check results
    found = storage_results.queue_get_by_id([queue_id])
    assert len(found) == 1
    assert found[0]["status"] == "COMPLETE"
    res = storage_results.get_results_by_task_id(queue_id)['data'][0]
    assert res['status'] == 'COMPLETE'

    # Check queue is empty
    r = storage_results.queue_get_next("test_manager")
    assert len(r) == 0


def test_storage_queue_duplicate(storage_results):

    result1 = storage_results.get_results()['data'][2]
    task1 = {
        # "hash_index": idx,
        "spec": {},
        "hooks": [("service", "123")],
        "tag": None,
        "base_result": ('results', result1['id'])
    }
    r = storage_results.queue_submit([task1])
    assert len(r["data"]) == 1
    queue_id = r["data"][0]

    # Put the first task in a waiting state
    r = storage_results.queue_get_next("test_manager")
    assert len(r) == 1

    # Change hooks, only one submission due to hash_index conflict
    task1["hooks"] = [("service", "456")]
    r = storage_results.queue_submit([task1])
    assert r["meta"]["n_inserted"] == 0

    # Pull out the data and check the hooks
    r = storage_results.queue_get_by_id([queue_id])
    hooks = r[0]["hooks"]
    assert len(hooks) == 2
    assert hooks[0][0] == "service"
    assert hooks[1][0] == "service"
    assert {"123", "456"} == {hooks[0][1], hooks[1][1]}

    # Cleanup
    r = storage_results.queue_mark_complete([queue_id])
    assert r == 1


def test_queue_submit_many_order(storage_results):

    results = storage_results.get_results()['data']

    task1 = {"base_result": ('results', results[3]['id'])}
    task2 = {"base_result": ('results', results[4]['id'])}
    task3 = {"base_result": ('results', results[5]['id'])}

    # Submit tasks
    ret = storage_results.queue_submit([task1, task2, task3])
    assert len(ret["data"]) == 3
    assert ret['meta']['n_inserted'] == 3

    # Get task
    r = storage_results.queue_get_next("test_manager", limit=1)
    assert len(r) == 1
    # will get the first submitted result first
    assert r[0]['base_result']['id'] == results[3]['id']

    # Todo: test more scenarios


# User testing


def test_user_duplicates(storage_socket):

    r = storage_socket.add_user("george", "shortpw")
    assert r is True

    # Duplicate should bounce
    r = storage_socket.add_user("george", "shortpw")
    assert r is False

    assert storage_socket.remove_user("george") is True

    assert storage_socket.remove_user("george") is False


def test_user_permissions_default(storage_socket):

    r = storage_socket.add_user("george", "shortpw")
    assert r is True

    # Verify correct permission
    assert storage_socket.verify_user("george", "shortpw", "read")[0] is True

    # Verify incorrect permission
    assert storage_socket.verify_user("george", "shortpw", "admin")[0] is False

    assert storage_socket.remove_user("george") is True


def test_user_permissions_admin(storage_socket):

    r = storage_socket.add_user("george", "shortpw", permissions=["read", "write", "compute", "admin"])
    assert r is True

    # Verify correct permissions
    assert storage_socket.verify_user("george", "shortpw", "read")[0] is True
    assert storage_socket.verify_user("george", "shortpw", "write")[0] is True
    assert storage_socket.verify_user("george", "shortpw", "compute")[0] is True
    assert storage_socket.verify_user("george", "shortpw", "admin")[0] is True

    assert storage_socket.remove_user("george") is True


def test_project_name(storage_socket):
    assert 'test' in storage_socket.get_project_name()



import datetime
import json
from collections.abc import Iterable

import bson
import mongoengine as db


class CustomDynamicDocument(db.DynamicDocument):
    """
    This class is serializable into standard json
    """

    def to_json_obj(self, with_id=True):
        """Removes object types like $date, _cls, and $oid
            Also, replaces _id with id
            Assumes one level of ReferenceFields which is suffice
        """

        data = json.loads(bson.json_util.dumps(self.to_mongo()))

        for key, value in data.items():
            if isinstance(value, dict) and len(value) == 1:
                (subkey, subvalue), = value.items()
                if subkey.startswith('$'):
                    data[key] = subvalue
            elif isinstance(value, Iterable) and '_ref' in value:
                data[key] = data[key]['_ref']
                data[key]['ref'] = data[key]['$ref']
                del data[key]['$ref']
                data[key]['id'] = str(data[key]['$id']['$oid'])
                del data[key]['$id']

        if with_id:
            data['id'] = data['_id']
        del data['_id']

        return data

    meta = {
        'abstract': True,
    }

class Collection(CustomDynamicDocument):
    """
        A collection of precomuted workflows such as datasets, ..

        This is a dynamic document, so it will accept any number of
        extra fields (expandable and uncontrolled schema)
    """

    collection = db.StringField(required=True)  # , choices=['dataset', '?'])
    name = db.StringField(required=True)  # Example 'water'

    meta = {
        'collection': 'collection',  # DB collection/table name
        'indexes': [
            {'fields': ('collection', 'name'), 'unique': True}
        ]
    }

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Molecule(CustomDynamicDocument):
    """
        The molecule DB collection is managed by pymongo, so far
    """

    name = db.StringField()
    symbols = db.ListField()
    molecular_formula = db.StringField()
    molecule_hash = db.StringField()
    geometry = db.ListField()
    real = db.ListField()
    fragments = db.DynamicField()

    def create_hash(self):
        """ TODO: create a special hash before saving"""
        return ''

    def save(self, *args, **kwargs):
        """Override save to add molecule_hash"""
        # self.molecule_hash = self.create_hash()

        return super(Molecule, self).save(*args, **kwargs)

    def __str__(self):
        return str(self.id)

    meta = {
        'collection': 'molecule',
        'indexes': [
            {'fields': ('molecule_hash',), 'unique': False},  # should almost be unique
            {'fields': ('molecular_formula',), 'unique': False}
        ]
    }

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Keywords(CustomDynamicDocument):
    """
        Keywords are unique for a specific program and name
    """

    # TODO: pull choices from const config
    program = db.StringField(required=True)  #, choices=['rdkit', 'psi4', 'geometric', 'torsiondrive'])
    hash_index = db.StringField(required=True)
    keywords = db.DynamicField()

    meta = {
        'indexes': [
            {'fields': ('program', 'hash_index'), 'unique': True}
        ]
    }

    def __str__(self):
        return str(self.id)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class BaseResult(CustomDynamicDocument):
    """
        Abstract Base class for Results and Procedures
    """

    # queue related
    task_id = db.StringField()  # ObjectId, reference task_queue but without validation
    status = db.StringField(required=True, choices=['COMPLETE', 'INCOMPLETE', 'ERROR'])

    meta = {
        'abstract': True,
        # 'allow_inheritance': True,
        'indexes': [
            'status'
        ]
    }

    def save(self, *args, **kwargs):
        """Override save to set defaults"""

        if not self.status:
            self.status = 'INCOMPLETE'

        return super(BaseResult, self).save(*args, **kwargs)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Result(BaseResult):
    """
        Hold the result of an atomic single calculation
    """

    # uniquely identifying a result
    program = db.StringField(required=True)  # example "rdkit", is it the same as program in keywords?
    driver = db.StringField(required=True)  # example "gradient"
    method = db.StringField(required=True)  # example "uff"
    basis = db.StringField()
    molecule = db.ReferenceField(Molecule, required=True)   # todo: or LazyReferenceField if only ID is needed?
    # options = db.ReferenceField(Keywords)  # ** has to be a FK or empty, can't be a string
    keywords = db.StringField()

    # output related
    properties = db.DynamicField()  # accept any, no validation
    return_result = db.DynamicField()  # better performance than db.ListField(db.FloatField())
    provenance = db.DynamicField()  # or an Embedded Documents with a structure?

    schema_name = db.StringField()  # default="qc_ret_data_output"??
    schema_version = db.IntField()  # or String?

    meta = {
        'collection': 'result',
        'indexes': [
           {'fields': ('program', 'driver', 'method', 'basis',
                       'molecule', 'keywords'), 'unique': True},
        ]
    }

    # not used yet
    # or  use pre_save
    def _save(self, *args, **kwargs):
        """Override save to handle keywords"""

        if not isinstance(self.keywords, Keywords):
            # self.keywords = Keywords.objects(program=self.program, option_name='default')\
            #     .modify(upsert=True, new=True, option_name='default')
            self.keywords = Keywords.objects(program=self.program, option_name='default').first()
            if not self.keywords:
                self.keywords = Keywords(program=self.program, option_name='default').save()
                # self.keywords.save()

        return super(Result, self).save(*args, **kwargs)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Procedure(BaseResult):
    """
        A procedure is a group of related results applied to a list of molecules
    """

    procedure = db.StringField(required=True)
                                    # choices=['optimization', 'torsiondrive'])
    # Todo: change name to be different from results program
    program = db.StringField(required=True)  # example: 'Geometric'
    keywords = db.DynamicField()  # keywords of the procedure

    hash_index = db.StringField()
    qc_meta = db.DynamicField()  # --> all inside results except mol

    meta = {
        'collection': 'procedure',
        'allow_inheritance': True,
        'indexes': [
            # TODO: needs a unique index, + molecule?
            {'fields': ('procedure', 'program'), 'unique': False},  # TODO: check
            {'fields': ('hash_index',), 'unique': False}  # used in queries
        ]
    }

# ================== Types of Procedures ================== #


class OptimizationProcedure(Procedure):
    """
        An Optimization  procedure
    """

    procedure = db.StringField(default='optimization', required=True)

    initial_molecule = db.ReferenceField(Molecule)  # always load with select_related
    final_molecule = db.ReferenceField(Molecule)

    # output
    # trajectory = db.ListField(Result)


class TorsiondriveProcedure(Procedure):
    """
        An torsion drive  procedure
    """

    procedure = db.StringField(default='torsiondrive', required=True)

    # TODO: add more fields

    meta = {
        'indexes': [
        ]
    }

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Spec(db.DynamicEmbeddedDocument):
    """ The spec of a task in the queue
        This is an embedded document, meaning that it will be embedded
        in the task_queue collection and won't be stored as a seperate
        collection/table --> for faster parsing
    """

    function = db.StringField()
    args = db.DynamicField()    # fast, can take any structure
    kwargs = db.DynamicField()


class TaskQueue(CustomDynamicDocument):
    """A queue of tasks corresponding to a procedure

       Notes: don't sort query results without having the index sorted
              will impact the performce
    """

    # spec = db.EmbeddedDocumentField(Spec, default=Spec)
    spec = db.DynamicField()

    # others
    hooks = db.ListField(db.DynamicField())  # ??
    tag = db.StringField(default=None)
    parser = db.StringField(default='')
    status = db.StringField(default='WAITING',
                            choices=['RUNNING', 'WAITING', 'ERROR', 'COMPLETE'])
    manager = db.StringField(default=None)

    created_on = db.DateTimeField(required=True, default=datetime.datetime.utcnow)
    modified_on = db.DateTimeField(required=True, default=datetime.datetime.utcnow)

    # can reference Results or any Procedure
    base_result = db.GenericLazyReferenceField(dbref=True)  # use res.id and res.document_type (class)

    meta = {
        'indexes': [
            'created_on',
            'status',
            'manager',
            # {'fields': ("status", "tag", "hash_index"), 'unique': False}
            {'fields': ('base_result',), 'unique': True},  # new
            # {'fields': ('hash_index',), 'unique': True}

        ]
        # 'indexes': [
        #         '$function', # text index, not needed
        #         '#function', # hash index
        #         ('title', '-rating'),  # rating is descending, direction only for multi-indices
        #     {
        #       'fields': ('spec.function', 'tag'),
        #       'unique': True
        #     }
        # ]
    }

    def save(self, *args, **kwargs):
        """Override save to update modified_on"""
        self.modified_on = datetime.datetime.utcnow()

        return super(TaskQueue, self).save(*args, **kwargs)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class ServiceQueue(CustomDynamicDocument):

    status = db.StringField()
    tag = db.StringField(default=None)
    hash_index = db.StringField(required=True)
    # procedure = db.ReferenceField()

    meta = {
        'indexes': [
            'status',
            {'fields': ("status", "tag", "hash_index"), 'unique': False},
            # {'fields': ('hash_index',), 'unique': True}

        ]
    }
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class User(CustomDynamicDocument):

    username = db.StringField(required=True, unique=True)
    password = db.BinaryField(required=True)
    permissions = db.ListField()

    meta = {
        'indexes': ['username']
    }


class QueueManager(CustomDynamicDocument):
    """
    """

    name = db.StringField(unique=True)
    cluster = db.StringField()
    hostname = db.StringField()
    uuid = db.StringField()
    tag = db.StringField()

    # counts
    completed = db.IntField(default=0)
    submitted = db.IntField(default=0)
    failures = db.IntField(default=0)
    returned = db.IntField(default=0)

    status = db.StringField(default='INACTIVE', choices=['ACTIVE', 'INACTIVE'])

    created_on = db.DateTimeField(required=True, default=datetime.datetime.utcnow)
    modified_on = db.DateTimeField(required=True, default=datetime.datetime.utcnow)

    meta = {
        'indexes': ['status', 'name', 'modified_on']
    }

    def save(self, *args, **kwargs):
        """Override save to update modified_on"""
        self.modified_on = datetime.datetime.utcnow()

        return super(QueueManager, self).save(*args, **kwargs)



"""
Assists in grabbing the requisite schema
"""

import copy

from .definitions_schema import get_definition
from .molecule_schema import molecule_schema
from .options_schema import options_schema

__all__ = [
    "get_schema", "get_table_indices", "get_schema_keys", "get_hash_fields", "format_result_indices"
]

_schemas = {}

# Add in molecule
for req in molecule_schema["required_definitions"]:
    molecule_schema["definitions"][req] = get_definition(req)

_schemas["molecule"] = molecule_schema
_schemas["options"] = options_schema

# Load molecule schema

# Collection and hash indices
_table_indices = {

    "collection": ("collection", "name"),
    "procedure": ("procedure", "program"),

    "molecule": ("molecule_hash", "molecular_formula"),
    "result": ("molecule", "program", "driver", "method", "basis", "options"),  # ** Renamed molecule_id
    "options": ("program", "name"),

    # "task_queue": ("status", "tag", "hash_index"),
    "task_queue": ("status", "tag", "base_result"),  # updated
    "service_queue": ("status", "tag", "hash_index"),
}  # yapf: disable


def get_hash_fields(name):
    if name not in _schemas:
        raise KeyError("Schema name {} not found.".format(name))
    return copy.deepcopy(_schemas[name]["hash_fields"])


def get_table_indices(name):
    if name not in _table_indices:
        raise KeyError("Indices for {} not found.".format(name))
    return _table_indices[name]


def format_result_indices(data, program=None):
    if program is None:
        program = data["program"]
    return program, data["molecule"], data["driver"], data["method"], data["basis"], data["options"]


def get_schema(name):
    if name not in _schemas:
        raise KeyError("Schema name {} not found.".format(name))
    return copy.deepcopy(_schemas)


def get_schema_keys(name):
    if name not in _schemas:
        raise KeyError("Schema name {} not found.".format(name))
    return _schemas[name]["properties"].keys()



"""
Tests the server collection compute capabilities.
"""

import pytest

import qcfractal.interface as portal
from qcfractal import testing
from qcfractal.testing import fractal_compute_server


@testing.using_psi4
def test_dataset_compute_gradient(fractal_compute_server):
    client = portal.FractalClient(fractal_compute_server)

    # Build a dataset
    ds = portal.collections.Dataset("ds_energy", client, default_program="psi4", default_driver="gradient")

    local = {"gradient": [0.03, 0, 0.02, -0.02, 0, -0.03]}
    ds.add_entry("He1", portal.Molecule.from_data("He -1 0 0\n--\nHe 0 0 1"), local_results=local)
    ds.add_entry("He2", portal.Molecule.from_data("He -1.1 0 0\n--\nHe 0 0 1.1"), local_results=local)

    # Compute
    ds.save()
    ds.compute("HF", "sto-3g")
    fractal_compute_server.await_results()

    ds.query("HF", "sto-3g", as_array=True)
    ds.query("gradient", None, local_results=True, as_array=True)

    stats = ds.statistics("MUE", "HF/sto-3g", "gradient")
    assert pytest.approx(stats.mean()) == 0.00984176986312362


def test_reactiondataset_check_state(fractal_compute_server):
    client = portal.FractalClient(fractal_compute_server)
    ds = portal.collections.ReactionDataset("check_state", client, ds_type="ie", default_program="rdkit")
    ds.add_ie_rxn("He1", portal.Molecule.from_data("He -3 0 0\n--\nHe 0 0 2"))

    with pytest.raises(ValueError):
        ds.compute("SCF", "STO-3G")

    with pytest.raises(ValueError):
        ds.query("SCF", "STO-3G")

    ds.save()
    assert ds.query("SCF", "STO-3G")

    ds.add_keywords("default", portal.models.KeywordSet(program="psi4", values={"a": 5}))

    with pytest.raises(ValueError):
        ds.query("SCF", "STO-3G")

    ds.save()
    assert ds.query("SCF", "STO-3G")


@testing.using_psi4
def test_compute_reactiondataset_regression(fractal_compute_server):
    """
    Tests an entire server and interaction energy dataset run
    """

    client = portal.FractalClient(fractal_compute_server)
    ds_name = "He_PES"
    ds = portal.collections.ReactionDataset(ds_name, client, ds_type="ie")

    # Add two helium dimers to the DB at 4 and 8 bohr
    He1 = portal.Molecule.from_data([[2, 0, 0, -2], [2, 0, 0, 2]], dtype="numpy", units="bohr", frags=[1])
    ds.add_ie_rxn("He1", He1, attributes={"r": 4}, reaction_results={"default": {"Benchmark": 0.0009608501557}})

    # Save the DB and re-acquire via classmethod
    r = ds.save()
    ds = portal.collections.ReactionDataset.from_server(client, ds_name)
    ds.set_default_program("psi4")
    assert "ReactionDataset(" in str(ds)

    # Test collection lists
    ret = client.list_collections()
    assert ds_name in ret["reactiondataset"]

    ret = client.list_collections("reactiondataset")
    assert ds_name in ret

    He2 = portal.Molecule.from_data([[2, 0, 0, -4], [2, 0, 0, 4]], dtype="numpy", units="bohr", frags=[1])
    ds.add_ie_rxn("He2", He2, attributes={"r": 4}, reaction_results={"default": {"Benchmark": -0.00001098794749}})

    # Save the DB and overwrite the result, reacquire via client
    r = ds.save(overwrite=True)
    ds = client.get_collection("reactiondataset", ds_name)

    # Compute SCF/sto-3g
    ret = ds.compute("SCF", "STO-3G")
    assert len(ret.submitted) == 3
    fractal_compute_server.await_results()

    # Query computed results
    assert ds.query("SCF", "STO-3G")
    assert pytest.approx(0.6024530476071095, 1.e-5) == ds.df.loc["He1", "SCF/STO-3G"]
    assert pytest.approx(-0.006895035942673289, 1.e-5) == ds.df.loc["He2", "SCF/STO-3G"]

    # Check results
    assert ds.query("Benchmark", "", reaction_results=True)
    assert pytest.approx(0.00024477933196125805, 1.e-5) == ds.statistics("MUE", "SCF/STO-3G")

    assert isinstance(ds.to_json(), dict)


@testing.using_psi4
def test_compute_reactiondataset_keywords(fractal_compute_server):

    client = portal.FractalClient(fractal_compute_server)

    mol1 = portal.Molecule.from_data("He 0 0 -1.1\n--\nHe 0 0 1.1")

    # Build a dataset
    ds = portal.collections.ReactionDataset("dataset_options", client, ds_type="ie")
    ds.set_default_program("Psi4")

    ds.add_ie_rxn("He2", mol1)
    ds.add_keywords("direct", portal.models.KeywordSet(program="psi4", values={"scf_type": "direct"}), default=True)
    ds.add_keywords("df", portal.models.KeywordSet(program="Psi4", values={"scf_type": "df"}))

    ds.save()

    # Compute, should default to direct options
    r = ds.compute("SCF", "STO-3G")
    fractal_compute_server.await_results()
    assert ds.query("SCF", "STO-3G")
    assert pytest.approx(0.39323818102293856, 1.e-5) == ds.df.loc["He2", "SCF/STO-3G"]

    r = ds.compute("SCF", "STO-3G", keywords="df")
    fractal_compute_server.await_results()
    assert ds.query("SCF", "STO-3G", keywords="df", prefix="df-")
    assert pytest.approx(0.38748602675524185, 1.e-5) == ds.df.loc["He2", "df-SCF/STO-3G"]


@testing.using_torsiondrive
@testing.using_geometric
@testing.using_rdkit
def test_compute_openffworkflow(fractal_compute_server):
    """
    Tests the openffworkflow collection
    """

    # Obtain a client and build a BioFragment
    client = portal.FractalClient(fractal_compute_server)

    openff_workflow_options = {
        # Blank Fragmenter options
        "enumerate_states": {},
        "enumerate_fragments": {},
        "torsiondrive_input": {},

        # TorsionDrive, Geometric, and QC options
        ""
        "torsiondrive_static_options": {
            "torsiondrive_meta": {},
            "optimization_meta": {
                "program": "geometric",
                "coordsys": "tric",
            },
            "qc_meta": {
                "driver": "gradient",
                "method": "UFF",
                "basis": "",
                "keywords": None,
                "program": "rdkit",
            }
        },
        "optimization_static_options": {
            "program": "geometric",
            "keywords": {
                "coordsys": "tric",
            },
            "qc_meta": {
                "driver": "gradient",
                "method": "UFF",
                "basis": "",
                "keywords": None,
                "program": "rdkit",
            }
        }
    }
    wf = portal.collections.OpenFFWorkflow("Workflow1", client=client, **openff_workflow_options)

    # # Add a fragment and wait for the compute
    hooh = portal.data.get_molecule("hooh.json")
    fragment_input = {
        "label1": {
            "type": "torsiondrive_input",
            "initial_molecule": hooh.json_dict(),
            "grid_spacing": [120],
            "dihedrals": [[0, 1, 2, 3]],
        },
    }
    wf.add_fragment("HOOH", fragment_input, provenance={})
    assert set(wf.list_fragments()) == {"HOOH"}
    fractal_compute_server.await_services(max_iter=5)

    final_energies = wf.list_final_energies()
    assert final_energies.keys() == {"HOOH"}
    assert final_energies["HOOH"].keys() == {"label1"}

    final_molecules = wf.list_final_molecules()
    assert final_molecules.keys() == {"HOOH"}
    assert final_molecules["HOOH"].keys() == {"label1"}

    optimization_input = {
        "label2": {
            "type": "optimization_input",
            "initial_molecule": hooh.json_dict(),
            "constraints": {
                'set': [{
                    "type": 'dihedral',
                    "indices": [0, 1, 2, 3],
                    "value": 0
                }]
            }
        }
    }

    wf.add_fragment("HOOH", optimization_input, provenance={})
    fractal_compute_server.await_services(max_iter=5)

    final_energies = wf.list_final_energies()
    assert final_energies["HOOH"].keys() == {"label1", "label2"}
    assert pytest.approx(0.00259754, 1.e-4) == final_energies["HOOH"]["label2"]

    final_molecules = wf.list_final_molecules()
    assert final_molecules.keys() == {"HOOH"}
    assert final_molecules["HOOH"].keys() == {"label1", "label2"}

    # Add a second fragment
    butane = portal.data.get_molecule("butane.json")
    butane_id = butane.identifiers.canonical_isomeric_explicit_hydrogen_mapped_smiles

    fragment_input = {
        "label1": {
            "type": "torsiondrive_input",
            "initial_molecule": butane.json_dict(),
            "grid_spacing": [90],
            "dihedrals": [[0, 2, 3, 1]],
        },
    }
    wf.add_fragment(butane_id, fragment_input, provenance={})
    assert set(wf.list_fragments()) == {butane_id, "HOOH"}

    final_energies = wf.list_final_energies()
    assert final_energies.keys() == {butane_id, "HOOH"}
    assert final_energies[butane_id].keys() == {"label1"}
    assert final_energies[butane_id]["label1"] is None


def test_generic_collection(fractal_compute_server):

    client = portal.FractalClient(fractal_compute_server)
    g = portal.collections.Generic("Generic1", client=client)

    # Check getters/savers
    g["hello"] = 5
    assert g["hello"] == 5

    # Copy the data over
    cp = g.get_data()
    assert cp.data["hello"] == 5
    cp.data["hello"] = 10

    # Make sure we did not change underlying data.
    assert g["hello"] == 5

    # Save semantics slightly wonky, double check this works
    g.save()
    g2 = portal.collections.Generic.from_server(client, "Generic1")
    assert g2["hello"] == 5


def test_missing_collection(fractal_compute_server):

    client = portal.FractalClient(fractal_compute_server)
    with pytest.raises(KeyError):
        client.get_collection("reactiondataset", "_waffles_")



"""
Utility functions for on-node procedures.
"""

import hashlib
import json

from .. import interface
from ..interface.models.common_models import ResultInput


def unpack_single_run_meta(storage, meta, molecules):
    """Transforms a metadata compute packet into an expanded
    QC Schema for multiple runs.

    Parameters
    ----------
    storage : DBSocket
        A live connection to the current database.
    meta : dict
        A JSON description of the metadata involved with the computation
    molecules : list of str, dict
        A list of molecule ID's or full JSON molecules associated with the run.

    Returns
    -------
    ret : tuple(dict, list)
        A dictionary of JSON representations with keys built in.
        The list is an array of any errors occurred

    Examples
    --------

    >>> meta = {
        "procedure": "single",
        "driver": "energy",
        "method": "HF",
        "basis": "sto-3g",
        "keywords": "default",
        "program": "psi4",
    }

    >>> molecules = [{"geometry": [0, 0, 0], "symbols" : ["He"]}]

    >>> unpack_single_run_meta(storage, meta, molecules)

    """

    # Get the required molecules
    raw_molecules_query = storage.get_add_molecules_mixed(molecules)

    # Pull out the needed keywords
    if meta["keywords"] is None:
        keyword_set = {}
    else:
        keyword_set = storage.get_add_keywords_mixed([meta["keywords"]])["data"][0]
        keyword_set = keyword_set["values"]

    # Create the "universal header"
    task_meta = json.dumps({
        "program": meta["program"],
        "driver": meta["driver"],
        "keywords": keyword_set,
        "model": {
            "method": meta["method"],
            "basis": meta["basis"]
        },
        "qcfractal_tags": {
            "program": meta["program"],
            "keywords": meta["keywords"]
        }
    })

    tasks = []
    for mol in raw_molecules_query["data"]:
        if mol is None:
            tasks.append(None)
            continue

        data = json.loads(task_meta)
        data["molecule"] = mol

        tasks.append(ResultInput(**data))

    return tasks, []


def parse_single_runs(storage, results):
    """Summary

    Parameters
    ----------
    storage : DBSocket
        A live connection to the current database.
    results : dict
        A (key, result) dictionary of the single return results.

    Returns
    -------

    Examples
    --------

    """

    # Get molecule ID's
    mols = {k: v["molecule"] for k, v in results.items()}
    mol_ret = storage.add_molecules(mols)["data"]

    for k, v in results.items():

        # Flatten data back out
        v["method"] = v["model"]["method"]
        v["basis"] = v["model"]["basis"]
        del v["model"]

        v["keywords"] = v["qcfractal_tags"]["keywords"]

        # Molecule should be by ID
        v["molecule"] = mol_ret[k]

        v["program"] = v["qcfractal_tags"]["program"]

        del v["qcfractal_tags"]
    return results


def single_run_hash(data, program=None):

    single_keys = interface.schema.format_result_indices(data, program=program)
    keys = {"procedure_type": "single", "single_key": single_keys}
    hash_index = hash_procedure_keys(keys)
    return keys, hash_index


def hash_procedure_keys(keys):
    m = hashlib.sha1()
    m.update(json.dumps(keys, sort_keys=True).encode("UTF-8"))
    return m.hexdigest()


def parse_hooks(rdata, rhooks):
    """Parses the hook data in a list of hooks
    TODO: this methos an has error, results is undefined

    Parameters
    ----------
    rdata : dict
        A {uid : results} dictionary of results to pull id's from
    rhooks : dict
        A {uid : hook} dictionary to apply the hooks too

    Returns
    -------
    TYPE
        Description
    """
    hook_data = []
    for k, hook in rhooks.items():

        # If no hooks skip it
        if len(hook) == 0:
            continue

        # Loop over hooks
        for h in hook:
            # Loop over individual commands
            for command in h["updates"]:
                # Custom commands
                if not isinstance(command[-1], str):
                    continue
                elif "$" not in command[-1]:
                    continue
                elif command[-1] == "$task_id":
                    command[-1] = results[k]["id"]
                elif command[-1] == "$hash_index":
                    command[-1] = results[k]["hash_index"]
                else:
                    raise KeyError("Hook command `{}` not understood.".format(command))

        hook_data.append(hook)
    return hook_data



"""
Tests the imports and exports of the Molecule object.
"""

import numpy as np
import pytest

from . import portal


def test_molecule_constructors():

    ### Water Dimer
    water_psi = portal.data.get_molecule("water_dimer_minima.psimol")
    ele = np.array([8, 1, 1, 8, 1, 1]).reshape(-1, 1)
    npwater = np.hstack((ele, water_psi.geometry))
    water_from_np = portal.Molecule.from_data(npwater, name="water dimer", dtype="numpy", frags=[3])

    assert water_psi.compare(water_psi, water_from_np)
    assert water_psi.get_molecular_formula() == "H4O2"

    # Check the JSON construct/deconstruct
    water_from_json = portal.Molecule(**water_psi.json_dict())
    assert water_psi.compare(water_psi, water_from_json)

    ### Neon Tetramer
    neon_from_psi = portal.data.get_molecule("neon_tetramer.psimol")
    ele = np.array([10, 10, 10, 10]).reshape(-1, 1)
    npneon = np.hstack((ele, neon_from_psi.geometry))
    neon_from_np = portal.Molecule.from_data(npneon, name="neon tetramer", dtype="numpy", frags=[1, 2, 3], units="bohr")

    assert neon_from_psi.compare(neon_from_psi, neon_from_np)

    # Check the JSON construct/deconstruct
    neon_from_json = portal.Molecule(**neon_from_psi.json_dict())
    assert neon_from_psi.compare(neon_from_psi, neon_from_json)
    assert neon_from_json.get_molecular_formula() == "Ne4"

    assert water_psi.compare(portal.Molecule.from_data(water_psi.to_string()))


def test_molecule_file_constructors():

    mol_psi = portal.data.get_molecule("helium_dimer.psimol")
    mol_json = portal.data.get_molecule("helium_dimer.json")
    mol_np = portal.data.get_molecule("helium_dimer.npy")

    assert mol_psi.compare(mol_json)
    assert mol_psi.compare(mol_np)
    assert mol_psi.get_molecular_formula() == "He2"


def test_water_minima_data():
    mol = portal.data.get_molecule("water_dimer_minima.psimol")

    assert len(str(mol)) == 662
    assert len(mol.to_string()) == 442

    assert sum(x == y for x, y in zip(mol.symbols, ['O', 'H', 'H', 'O', 'H', 'H'])) == mol.geometry.shape[0]
    assert mol.molecular_charge == 0
    assert mol.molecular_multiplicity == 1
    assert np.sum(mol.real) == mol.geometry.shape[0]
    assert np.allclose(mol.fragments, [[0, 1, 2], [3, 4, 5]])
    assert np.allclose(mol.fragment_charges, [0, 0])
    assert np.allclose(mol.fragment_multiplicities, [1, 1])
    assert hasattr(mol, "provenance")
    assert np.allclose(
        mol.geometry,
        [[2.81211080, 0.1255717, 0.], [3.48216664, -1.55439981, 0.], [1.00578203, -0.1092573, 0.],
         [-2.6821528, -0.12325075, 0.], [-3.27523824, 0.81341093, 1.43347255], [-3.27523824, 0.81341093, -1.43347255]])
    assert mol.get_hash() == "b41f1e38bc4be5482fcd1d4dd53ca7c65146ab91"


def test_water_minima_fragment():

    mol = portal.data.get_molecule("water_dimer_minima.psimol")

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)
    assert frag_0.get_hash() == "6d253a5a66eb68b611ab6bb0f15b55bbd3f6fe91"
    assert frag_1.get_hash() == "feb5c6127ca54d715b999c15ea1ea1772ada8c5d"

    frag_0_1 = mol.get_fragment(0, 1)
    frag_1_0 = mol.get_fragment(1, 0)

    assert mol.symbols[:3] == frag_0.symbols
    assert np.allclose(mol.masses[:3], frag_0.masses)

    assert mol.symbols == frag_0_1.symbols
    assert np.allclose(mol.geometry, frag_0_1.geometry)

    assert mol.symbols[3:] + mol.symbols[:3] == frag_1_0.symbols
    assert np.allclose(mol.masses[3:] + mol.masses[:3], frag_1_0.masses)


def test_pretty_print():

    mol = portal.data.get_molecule("water_dimer_minima.psimol")
    assert isinstance(mol.pretty_print(), str)


def test_to_string():

    mol = portal.data.get_molecule("water_dimer_minima.psimol")
    assert isinstance(mol.to_string(), str)


def test_water_orient():
    # These are identical molecules, should find the correct results

    mol = portal.data.get_molecule("water_dimer_stretch.psimol")
    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True)

    assert frag_0_1.get_hash() == frag_1_0.get_hash()

    mol = portal.data.get_molecule("water_dimer_stretch2.psimol")
    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.molecular_multiplicity == 1
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True)

    # Ghost fragments should prevent overlap
    assert frag_0_1.molecular_multiplicity == 1
    assert frag_0_1.get_hash() != frag_1_0.get_hash()


def test_molecule_errors():
    mol = portal.data.get_molecule("water_dimer_stretch.psimol")

    data = mol.json_dict()
    data["whatever"] = 5
    with pytest.raises(ValueError):
        portal.Molecule(**data)


def test_molecule_repeated_hashing():

    mol = portal.Molecule(**{
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
            1.73178198, 1.29095807, 1.03716028, 1.31566305, -0.007440200000000001, -0.28074722, -1.3143081, 0.00849608,
            -0.27416914, -1.7241109, -1.30793432, 1.02770172
        ]
    })

    h1 = mol.get_hash()
    assert mol.get_molecular_formula() == "H2O2"

    mol2 = portal.Molecule(**mol.json_dict(), orient=False)
    assert h1 == mol2.get_hash()

    mol3 = portal.Molecule(**mol2.json_dict(), orient=False)
    assert h1 == mol3.get_hash()



"""
Tests the interface portal adapter to the REST API
"""

import qcfractal.interface as portal
from qcfractal.testing import test_server
import pytest

# All tests should import test_server, but not use it
# Make PyTest aware that this module needs the server


def test_client_molecule(test_server):

    client = portal.FractalClient(test_server)

    water = portal.data.get_molecule("water_dimer_minima.psimol")

    # Test add
    ret = client.add_molecules({"water": water})

    # Test get
    get_mol = client.get_molecules(ret["water"], index="id")
    assert water.compare(get_mol[0])

    # Test molecular_formula get
    get_mol = client.get_molecules(["H4O2"], index="molecular_formula")
    assert water.compare(get_mol[0])


def test_client_options(test_server):

    client = portal.FractalClient(test_server)

    opt = portal.models.KeywordSet(program="psi4", values={"one": "fish", "two": "fish"})

    # Test add
    ret = client.add_keywords([opt])

    # Test get
    get_kw = client.get_keywords({'id': ret[0]})
    assert opt == get_kw[0]

    get_kw = client.get_keywords({"program": "psi4", "hash_index": opt.hash_index})
    assert opt == get_kw[0]


def test_client_duplicate_keywords(test_server):

    client = portal.FractalClient(test_server)

    opt1 = portal.models.KeywordSet(program="psi4", values={"key": 1})
    opt2 = portal.models.KeywordSet(program="psi4", values={"key": 2})
    opt3 = portal.models.KeywordSet(program="psi4", values={"key": 3})

    # Test add
    ret = client.add_keywords([opt1, opt1])
    assert len(ret) == 2
    assert ret[0] == ret[1]

    ret2 = client.add_keywords([opt1])
    assert len(ret2) == 1
    assert ret2[0] == ret[0]

    ret3 = client.add_keywords([opt2, opt1, opt3])
    assert len(ret3) == 3
    assert ret3[1] == ret[0]


def test_collection_portal(test_server):

    db = {"collection": "torsiondrive", "name": "Torsion123", "something": "else", "array": ["54321"]}

    client = portal.FractalClient(test_server)

    # Test add
    _ = client.add_collection(db)

    # Test get
    get_db = client.get_collection(db["collection"], db["name"], full_return=True)
    db_id = get_db.data[0].pop("id")

    assert db == get_db.data[0]

    # Test add w/o overwrite
    ret = client.add_collection(db, full_return=True)
    assert ret.meta.success is False

    # Test that client is smart enough to trap non-id'ed overwrites
    with pytest.raises(KeyError):
        _ = client.add_collection(db, overwrite=True)

    # Test that we cannot use a local key
    db['id'] = 'local'
    db['array'] = ["12345"]
    with pytest.raises(KeyError):
        _ = client.add_collection(db, overwrite=True)

    # Finally test that we can overwrite
    db['id'] = db_id
    _ = client.add_collection(db, overwrite=True)
    get_db = client.get_collection(db["collection"], db["name"], full_return=True)
    assert get_db.data[0]['array'] == ["12345"]



"""
Wraps geometric procedures
"""

import copy
import json
from typing import Dict, Set

import numpy as np

from qcfractal.extras import get_information
from qcfractal.interface.models.common_models import json_encoders
from qcfractal.interface.models.gridoptimization import GridOptimization
from .service_util import BaseService, TaskManager, expand_ndimensional_grid

__all__ = ["GridOptimizationService"]


class GridOptimizationService(BaseService):

    # Index info
    status: str = "READY"
    service: str = "gridoptimization"
    program: str = "qcfractal"
    procedure: str = "gridoptimization"

    # Output
    output: GridOptimization

    # Temporaries
    grid_optimizations: Dict[str, str] = {}
    seeds: Set[tuple] = set()
    complete: Set[tuple] = set()
    dimensions: tuple
    iteration: int = 0
    starting_grid: tuple
    final_energies = {}

    # Task helpers
    task_map: Dict[str, str] = {}
    task_manager: TaskManager = TaskManager()

    # Templates
    constraint_template: str
    optimization_template: str
    molecule_template: str

    class Config:
        json_encoders = json_encoders

    @classmethod
    def initialize_from_api(cls, storage_socket, service_input):

        # Build the results object
        input_dict = service_input.dict()
        input_dict["initial_molecule"] = input_dict["initial_molecule"]["id"]

        output = GridOptimization(
            **input_dict,
            provenance={
                "creator": "qcfractal",
                "version": get_information("version"),
                "routine": "qcfractal.services.gridoptimization"
            },
            final_energy_dict={},
            grid_optimizations={},
            starting_grid=[0])

        meta = {"output": output}

        # Remove identity info from molecule template
        molecule_template = copy.deepcopy(service_input.initial_molecule.json_dict())
        molecule_template.pop("id", None)
        molecule_template.pop("identifiers", None)
        meta["molecule_template"] = json.dumps(molecule_template)

        # Build dihedral template
        constraint_template = []
        for scan in output.gridoptimization_meta.scans:
            tmp = {"type": scan.type, "indices": scan.indices}
            constraint_template.append(tmp)

        meta["constraint_template"] = json.dumps(constraint_template)

        # Build optimization template
        meta["optimization_template"] = json.dumps({
            "meta": {
                "procedure": "optimization",
                "keywords": {"program": output.optimization_meta.program, "values": output.optimization_meta.dict()},
                "program": output.optimization_meta.program,
                "qc_meta": output.qc_meta.dict(),
                "tag": meta.pop("tag", None)
            },
        })

        # Move around geometric data
        meta["optimization_program"] = output.optimization_meta.program

        meta["hash_index"] = output.hash_index

        # Hard coded data, # TODO
        meta["dimensions"] = output.get_scan_dimensions()

        if output.gridoptimization_meta.starting_grid == "zero":
            meta["iteration"] = 0
            meta["starting_grid"] = (0 for x in meta["dimensions"])
        elif output.gridoptimization_meta.starting_grid == "relative":
            meta["iteration"] = 0
            starting_grid = []
            for scan in output.gridoptimization_meta.scans:

                # Find closest index
                m = service_input.initial_molecule.measure(scan.indices)
                idx = np.abs(np.array(scan.steps) - m).argmin()
                starting_grid.append(int(idx))

            meta["starting_grid"] = tuple(starting_grid)

        else:
            raise KeyError(
                "Unknown starting_grid configuration {}.".format(output.gridoptimization_meta.starting_grid))

        return cls(**meta, storage_socket=storage_socket)

    def iterate(self):

        self.status = "RUNNING"

        if self.iteration == 0:

            self.submit_optimization_tasks({
                self.output.serialize_key(self.starting_grid): self.output.initial_molecule
            })
            self.iteration = 1

            return False

        # Check if tasks are done
        if self.task_manager.done(self.storage_socket) is False:
            return False

        # Obtain complete tasks and figure out future tasks
        complete_tasks = self.task_manager.get_tasks(self.storage_socket)
        for k, v in complete_tasks.items():
            self.final_energies[k] = v["energies"][-1]

        # Build out nthe new set of seeds
        complete_seeds = set(tuple(json.loads(k)) for k in complete_tasks.keys())
        self.complete |= complete_seeds
        self.seeds = complete_seeds
        # print("Complete", self.complete)

        # Compute new points
        new_points_list = expand_ndimensional_grid(self.dimensions, self.seeds, self.complete)
        # print(new_points_list)

        # grid = np.zeros(self.dimensions, dtype=np.int)
        # for x in self.complete:
        #     grid[x] = 1
        # print(grid)

        next_tasks = {}
        for new_points in new_points_list:
            old = self.output.serialize_key(new_points[0])
            new = self.output.serialize_key(new_points[1])

            next_tasks[new] = complete_tasks[old]["final_molecule"]

        # All done
        if len(next_tasks) == 0:
            return self.finalize()

        self.submit_optimization_tasks(next_tasks)

        return False

    def submit_optimization_tasks(self, task_dict):

        new_tasks = {}

        for key, mol in task_dict.items():

            # Update molecule
            packet = json.loads(self.optimization_template)

            # Construct constraints
            constraints = json.loads(self.constraint_template)
            grid_values = self.output.get_scan_value(key)
            for con_num, k in enumerate(grid_values):
                constraints[con_num]["value"] = k
            packet["meta"]["keywords"]["values"]["constraints"] = {"set": constraints}

            # Build new molecule
            packet["data"] = [mol]

            new_tasks[key] = packet

        self.task_manager.submit_tasks(self.storage_socket, "optimization", new_tasks)

    def finalize(self):
        """
        Finishes adding data to the GridOptimization object
        """

        self.output.Config.allow_mutation = True
        self.output.success = True
        self.output.status = "COMPLETE"

        self.output.starting_grid = self.starting_grid
        self.output.grid_optimizations = self.grid_optimizations
        self.output.final_energy_dict = self.final_energies

        self.output.Config.allow_mutation = False
        return self.output



"""
Manipulates available services.
"""

from .gridoptimization_service import GridOptimizationService
from .torsiondrive_service import TorsionDriveService

__all__ = ["initializer", "build"]


def _service_chooser(name):
    """
    Choose the correct service
    """
    name = name.lower()
    if name == "torsiondrive":
        return TorsionDriveService
    elif name == "gridoptimization":
        return GridOptimizationService
    else:
        raise KeyError("Name {} not recognized.".format(name.title()))


def initializer(storage_socket, service_input):
    """Initializes a service from a API call

    Parameters
    ----------
    name : str
        Description
    storage_socket : StorageSocket
        A StorageSocket to the currently active database
    meta : dict
        A JSON blob with the required keys for the service.
    molecule : dict
        A JSON Molecule object

    Returns
    -------
    Service
        Returns an instantiated service

    """
    name = service_input.procedure
    return _service_chooser(name).initialize_from_api(storage_socket, service_input)


def build(storage_socket, data):
    """Initializes a service from a JSON blob

    Parameters
    ----------
    storage_socket : StorageSocket
        A StorageSocket to the currently active database
    data : dict
        The associated JSON blob with the service

    Returns
    -------
    Service
        Returns an instantiated service

    """
    name = data["service"]
    return _service_chooser(name)(**data, storage_socket=storage_socket)



"""
The json-schema for the Options definition
"""

options_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "properties": {
        "name": {
            "description": "The name of the options definition.",
            "type": "string",
        },
        "program": {
            "description": "The name of the program these options are associated with.",
            "type": "string",
        },
    },
    "required": ["program", "name"],
    "description": "The physical cartesian representation of the molecular system",
    "definitions": {},
    "additionalProperties": True,
}



"""Mongo QCDB Fragment object and helpers
"""
import itertools as it

import numpy as np
import pandas as pd

# from .. import client
from .. import constants
from .. import dict_utils
from .. import molecule
from .. import statistics
from .collection import Collection
from .collection_utils import nCr


class Fragment(Collection):
    """
    This is a QCA Database class.

    Attributes
    ----------
    client : client.FractalClient
        A optional server portal to connect the database
    data : dict
        JSON representation of the database backbone
    df : pd.DataFrame
        The underlying dataframe for the Database object
    torsion_index : pd.Index
        The unrolled results of all torsion scans in tied to the Fragment
    """

    def __init__(self, name, client=None):
        """
        Initializer for the Fragment object. If no Portal is supplied or the database name
        is not present on the server that the Portal is connected to a blank database will be
        created.

        Parameters
        ----------
        name : str
            The name of the Fragment
        client : client.FractalClient, optional
            A Portal client to connect to a server
        """
        super().__init__(name, client=client)

        # Internal data
        self.torsion_index = pd.DataFrame()

        # Needs to be input client, not self.client
        if client is not None:

            self.df = pd.DataFrame(index=self.get_index())

            # Unroll the index
            tmp_index = []
            for torsion in self.data["torsions"]:
                name = torsion["name"]




                for stoich_name in list(rxn["stoichiometry"]):
                    for mol_hash, coef in rxn["stoichiometry"][stoich_name].items():
                        tmp_index.append([name, stoich_name, mol_hash, coef])

            self.rxn_index = pd.DataFrame(tmp_index, columns=["name", "stoichiometry", "molecule_id", "coefficient"])

        # If we making a new database we may need new hashes and json objects
        self._new_molecule_jsons = {}

    def _init_collection_data(self, additional_args):
        return {"torsions": []}


    ### LINE OF THINGS I HAVE TOUCHED






    def _pre_save_prep(self, client):

        mol_ret = client.add_molecules(self._new_molecule_jsons)

        # Update internal molecule UUID's to servers UUID's
        self.data["reactions"] = dict_utils.replace_dict_keys(self.data["reactions"], mol_ret)
        self._new_molecule_jsons = {}

    def _unroll_query(self, keys, stoich, field="result_result"):
        """Unrolls a complex query into a "flat" query for the server object

        Parameters
        ----------
        keys : dict
            Server query fields
        stoich : str
            The stoichiometry to access for the query (default/cp/cp3/etc)
        field : str, optional
            The results field to query on

        Returns
        -------
        ret : pd.DataFrame
            A DataFrame representation of the unrolled query
        """
        tmp_idx = self.rxn_index[self.rxn_index["stoichiometry"] == stoich].copy()
        tmp_idx = tmp_idx.reset_index(drop=True)

        # There could be duplicates so take the unique and save the map
        umols, uidx = np.unique(tmp_idx["molecule_id"], return_index=True)

        # Evaluate the overall dataframe
        query_keys = {k: v for k, v in keys.items()}
        query_keys["molecule_id"] = list(umols)
        query_keys["projection"] = {field: True, "molecule_id": True}
        values = pd.DataFrame(self.client.get_results(**query_keys))

        # Join on molecule hash
        tmp_idx = tmp_idx.merge(values, how="left", on="molecule_id")

        # Apply stoich values
        for col in values.columns:
            if col == "molecule_id": continue
            tmp_idx[col] *= tmp_idx["coefficient"]
        tmp_idx = tmp_idx.drop(['stoichiometry', 'molecule_id', 'coefficient'], axis=1)

        # If *any* value is null in the stoich sum, the whole thing should be Null. Pandas is being too clever
        null_mask = tmp_idx.copy()
        null_mask[field] = null_mask[field].isnull()
        null_mask = null_mask.groupby(["name"]).sum() != False

        tmp_idx = tmp_idx.groupby(["name"]).sum()
        tmp_idx[null_mask] = np.nan

        return tmp_idx

    def query(self,
              method,
              basis,
              driver="energy",
              options="default",
              program="psi4",
              stoich="default",
              prefix="",
              postfix="",
              reaction_results=False,
              scale="kcal",
              field="return_result",
              ignore_db_type=False):
        """
        Queries the local Portal for the requested keys and stoichiometry.

        Parameters
        ----------
        method : str
            The computational method to query on (B3LYP)
        basis : str
            The computational basis query on (6-31G)
        driver : str, optional
            Search within energy, gradient, etc computations
        options : str, optional
            The option token desired
        program : str, optional
            The program to query on
        stoich : str
            The given stoichiometry to compute.
        prefix : str
            A prefix given to the resulting column names.
        postfix : str
            A postfix given to the resulting column names.
        reaction_results : bool
            Toggles a search between the Mongo Pages and the Databases's reaction_results field.
        scale : str, double
            All units are based in hartree, the default scaling is to kcal/mol.
        field : str, optional
            The result field to query on
        ignore_db_type : bool
            Override of IE for RXN db types.


        Returns
        -------
        success : bool
            Returns True if the requested query was successful or not.

        Notes
        -----


        Examples
        --------

        db.query("B3LYP", "aug-cc-pVDZ", stoich="cp", prefix="cp-")

        """

        if not reaction_results and (self.client is None):
            raise AttributeError("DataBase: FractalClient was not set.")

        query_keys = {
            "method": method.lower(),
            "basis": basis.lower(),
            "driver": driver.lower(),
            "options": options.lower(),
            "program": program.lower(),
        }
        # # If reaction results
        if reaction_results:
            tmp_idx = pd.Series(index=self.df.index)
            for rxn in self.data["reactions"]:
                try:
                    tmp_idx.ix[rxn["name"]] = rxn["reaction_results"][stoich][method]
                except KeyError:
                    pass

            # Convert to numeric
            tmp_idx = pd.to_numeric(tmp_idx, errors='ignore')
            tmp_idx *= constants.get_scale(scale)

            self.df[prefix + method + postfix] = tmp_idx
            return True

        # if self.data["db_type"].lower() == "ie":
        #     _ie_helper(..)

        if (not ignore_db_type) and (self.data["db_type"].lower() == "ie"):
            monomer_stoich = ''.join([x for x in stoich if not x.isdigit()]) + '1'
            tmp_idx_complex = self._unroll_query(query_keys, stoich, field=field)
            tmp_idx_monomers = self._unroll_query(query_keys, monomer_stoich, field=field)

            # Combine
            tmp_idx = tmp_idx_complex - tmp_idx_monomers

        else:
            tmp_idx = self._unroll_query(query_keys, stoich, field=field)
        tmp_idx.columns = [prefix + method + '/' + basis + postfix for x in tmp_idx.columns]

        # scale
        tmp_idx = tmp_idx.apply(lambda x: pd.to_numeric(x, errors='ignore'))
        tmp_idx[tmp_idx.select_dtypes(include=['number']).columns] *= constants.get_scale(scale)

        # Apply to df
        self.df[tmp_idx.columns] = tmp_idx

        return True

    def compute(self,
                method,
                basis,
                driver="energy",
                stoich="default",
                options="default",
                program="psi4",
                ignore_db_type=False):
        """Executes a computational method for all reactions in the Database.
        Previously completed computations are not repeated.

        Parameters
        ----------
        method : str
            The computational method to compute (B3LYP)
        basis : str
            The computational basis to compute (6-31G)
        driver : str, optional
            The type of computation to run (energy, gradient, etc)
        stoich : str, optional
            The stoichiometry of the requested compute (cp/nocp/etc)
        options : str, optional
            The options token for the requested compute
        program : str, optional
            The underlying QC program
        ignore_db_type : bool, optional
            Optionally only compute the "default" geometry

        Returns
        -------
        ret : dict
            A dictionary of the keys for all requested computations
        """
        if self.client is None:
            raise AttributeError("DataBase: Compute: Client was not set.")

        # Figure out molecules that we need
        if (not ignore_db_type) and (self.data["db_type"].lower() == "ie"):
            monomer_stoich = ''.join([x for x in stoich if not x.isdigit()]) + '1'
            tmp_monomer = self.rxn_index[self.rxn_index["stoichiometry"] == monomer_stoich].copy()
            tmp_complex = self.rxn_index[self.rxn_index["stoichiometry"] == stoich].copy()
            tmp_idx = pd.concat((tmp_monomer, tmp_complex), axis=0)
        else:
            tmp_idx = self.rxn_index[self.rxn_index["stoichiometry"] == stoich].copy()

        tmp_idx = tmp_idx.reset_index(drop=True)

        # There could be duplicates so take the unique and save the map
        umols, uidx = np.unique(tmp_idx["molecule_id"], return_index=True)

        complete_values = self.client.get_results(
            molecule_id=list(umols), driver=driver, options=options, program=program, method=method, basis=basis)

        if len(complete_values):
            raise KeyError("Completed expressions not yet implemented")
        # mask = pd.isnull(values)
        # compute_list = []
        # for idx, row in pd.isnull(values).iterrows():

        #     for method in values.columns[row]:
        #         tmp = {}
        #         tmp["molecule_id"] = idx
        #         tmp["modelchem"] = method
        #         for k, v in other_fields.items():
        #             tmp[k] = v
        #         compute_list.append(tmp)
        compute_list = list(umols)

        ret = self.client.add_compute(program, method.lower(), basis.lower(), driver, options, compute_list)

        return ret

    def get_index(self):
        """
        Returns the current index of the database.

        Returns
        -------
        ret : list of str
            The names of all reactions in the database
        """
        return [x["name"] for x in self.data["torsions"]]

    def get_rxn(self, name):
        """
        Returns the JSON object of a specific reaction.

        Parameters
        ----------
        name : str
            The name of the reaction to query

        Returns
        -------
        ret : dict
            The JSON representation of the reaction

        """

        found = []
        for num, x in enumerate(self.data["reactions"]):
            if x["name"] == name:
                found.append(num)

        if len(found) == 0:
            raise KeyError("Database:get_rxn: Reaction name '{}' not found.".format(name))

        if len(found) > 1:
            raise KeyError("Database:get_rxn: Multiple reactions of name '{}' found. Database failure.".format(name))

        return self.data["reactions"][found[0]]

    # Statistical quantities
    def statistics(self, stype, value, bench="Benchmark"):
        """Summary

        Parameters
        ----------
        stype : str
            The type of statistic in question
        value : str
            The method string to compare
        bench : str, optional
            The benchmark method for the comparison

        Returns
        -------
        ret : pd.DataFrame, pd.Series, float
            Returns a DataFrame, Series, or float with the requested statistics depending on input.
        """
        return statistics.wrap_statistics(stype, self.df, value, bench)

    # Visualization
    def ternary(self, cvals=None):
        """Plots a ternary diagram of the DataBase if available

        Parameters
        ----------
        cvals : None, optional
            Description

        """
        raise Exception("MPL not avail")


#        return visualization.Ternary2D(self.df, cvals=cvals)

# Adders

    def parse_stoichiometry(self, stoichiometry):
        """
        Parses a stiochiometry list.

        Parameters
        ----------
        stoichiometry : list
            A list of tuples describing the stoichiometry.

        Returns
        -------
        stoich : list
            A list of formatted tuples describing the stoichiometry for use in a MongoDB.

        Notes
        -----
        This function attempts to convert the molecule into its correspond hash. The following will happen depending on the form of the Molecule.
            - Molecule hash - Used directly in the stoichiometry.
            - Molecule class - Hash is obtained and the molecule will be added to the database upon saving.
            - Molecule string - Molecule will be converted to a Molecule class and the same process as the above will occur.


        Examples
        --------


        """

        ret = {}

        mol_hashes = []
        mol_values = []

        for line in stoichiometry:
            if len(line) != 2:
                raise KeyError("Database: Parse stoichiometry: passed in as a list must of key : value type")

            # Get the values
            try:
                mol_values.append(float(line[1]))
            except:
                raise TypeError("Database: Parse stoichiometry: must be able to cast second value must be as float.")

            # What kind of molecule is it?
            mol = line[0]

            # This is a molecule hash, should be in the database
            if isinstance(mol, str) and (len(mol) == 40):
                molecule_hash = mol

            elif isinstance(mol, str):
                qcdb_mol = molecule.Molecule(mol)

                molecule_hash = qcdb_mol.get_hash()

                if molecule_hash not in list(self._new_molecule_jsons):
                    self._new_molecule_jsons[molecule_hash] = qcdb_mol.to_json()

            elif isinstance(mol, molecule.Molecule):
                molecule_hash = mol.get_hash()

                if molecule_hash not in list(self._new_molecule_jsons):
                    self._new_molecule_jsons[molecule_hash] = mol.to_json()

            else:
                raise TypeError(
                    "Database: Parse stoichiometry: first value must either be a molecule hash, "
                    "a molecule str, or a Molecule class."
                )

            mol_hashes.append(molecule_hash)

        # Sum together the coefficients of duplicates
        ret = {}
        for mol, coef in zip(mol_hashes, mol_values):
            if mol in list(ret):
                ret[mol] += coef
            else:
                ret[mol] = coef

        return ret

    def add_rxn(self, name, stoichiometry, reaction_results=None, attributes=None, other_fields=None):
        """
        Adds a reaction to a database object.

        Parameters
        ----------
        name : str
            Name of the reaction.
        stoichiometry : list or dict
            Either a list or dictionary of lists
        reaction_results : None, optional
            A dictionary of the computed total interaction energy results
        attributes : None, optional
            A dictionary of attributes to assign to the reaction
        other_fields : None, optional
            A dictionary of additional user defined fields to add to the reaction entry

        Notes
        -----

        Examples
        --------

        Returns
        -------
        ret : dict
            A complete JSON specification of the reaction


        """
        if reaction_results is None:
            reaction_results = {}
        if attributes is None:
            attributes = {}
        if other_fields is None:
            other_fields = {}
        rxn = {"name": name}

        # Set name
        if name in self.get_index():
            raise KeyError(
                "Database: Name '{}' already exists. "
                "Please either delete this entry or call the update function.".format(name))

        # Set stoich
        if isinstance(stoichiometry, dict):
            rxn["stoichiometry"] = {}

            if "default" not in list(stoichiometry):
                raise KeyError("Database:add_rxn: Stoichiometry dict must have a 'default' key.")

            for k, v in stoichiometry.items():
                rxn["stoichiometry"][k] = self.parse_stoichiometry(v)

        elif isinstance(stoichiometry, (tuple, list)):
            rxn["stoichiometry"] = {}
            rxn["stoichiometry"]["default"] = self.parse_stoichiometry(stoichiometry)
        else:
            raise TypeError("Database:add_rxn: Type of stoichiometry input was not recognized:",
                            type(stoichiometry))

        # Set attributes
        if not isinstance(attributes, dict):
            raise TypeError("Database:add_rxn: attributes must be a dictionary, not '{}'".format(type(attributes)))

        rxn["attributes"] = attributes

        if not isinstance(other_fields, dict):
            raise TypeError("Database:add_rxn: other_fields must be a dictionary, not '{}'".format(type(attributes)))

        for k, v in other_fields.items():
            rxn[k] = v

        if "default" in list(reaction_results):
            rxn["reaction_results"] = reaction_results
        elif isinstance(reaction_results, dict):
            rxn["reaction_results"] = {}
            rxn["reaction_results"]["default"] = reaction_results
        else:
            raise TypeError("Passed in reaction_results not understood.")

        self.data["reactions"].append(rxn)

        return rxn

    def add_ie_rxn(self, name, mol, **kwargs):
        """Add a interaction energy reaction entry to the database. Automatically
        builds CP and no-CP reactions for the fragmented molecule.

        Parameters
        ----------
        name : str
            The name of the reaction
        mol : Molecule
            A molecule with multiple fragments
        **kwargs
            Additional kwargs to pass into `build_id_fragments`.

        Returns
        -------
        ret : dict
            A JSON representation of the new reaction.
        """
        reaction_results = kwargs.pop("reaction_results", {})
        attributes = kwargs.pop("attributes", {})
        other_fields = kwargs.pop("other_fields", {})

        stoichiometry = self.build_ie_fragments(mol, name=name, **kwargs)
        return self.add_rxn(
            name, stoichiometry, reaction_results=reaction_results, attributes=attributes, other_fields=other_fields)

    @staticmethod
    def build_ie_fragments(mol, **kwargs):
        """
        Build the stoichiometry for an Interaction Energy.

        Parameters
        ----------
        mol : Molecule class or str
            Molecule to fragment.
        do_default : bool
            Create the default (noCP) stoichiometry.
        do_cp : bool
            Create the counterpoise (CP) corrected stoichiometry.
        do_vmfc : bool
            Create the Valiron-Mayer Function Counterpoise (VMFC) corrected stoichiometry.
        max_nbody : int
            What is the maximum fragment level built, if zero defaults to the maximum number of fragments.

        Notes
        -----

        Examples
        --------

        Returns
        -------
        ret : dict
            A JSON representation of the fragmented molecule.

        """

        do_default = kwargs.pop("do_default", True)
        do_cp = kwargs.pop("do_cp", True)
        do_vmfc = kwargs.pop("do_vmfc", False)
        max_nbody = kwargs.pop("max_nbody", 0)

        if not isinstance(mol, molecule.Molecule):

            mol = molecule.Molecule(mol, **kwargs)

        ret = {}

        max_frag = len(mol.fragments)
        if max_nbody == 0:
            max_nbody = max_frag

        if max_nbody < 2:
            raise AttributeError("Database:build_ie_fragments: Molecule must have at least two fragments.")

        # Build some info
        fragment_range = list(range(max_frag))

        # Loop over the bodis
        for nbody in range(1, max_nbody):
            nocp_tmp = []
            cp_tmp = []
            for k in range(1, nbody + 1):
                take_nk = nCr(max_frag - k - 1, nbody - k)
                sign = ((-1)**(nbody - k))
                coef = take_nk * sign
                for frag in it.combinations(fragment_range, k):
                    if do_default:
                        nocp_tmp.append((mol.get_fragment(frag), coef))
                    if do_cp:
                        ghost = list(set(fragment_range) - set(frag))
                        cp_tmp.append((mol.get_fragment(frag, ghost), coef))

            if do_default:
                ret["default" + str(nbody)] = nocp_tmp

            if do_cp:
                ret["cp" + str(nbody)] = cp_tmp

        # VMFC is a special beast
        if do_vmfc:
            raise KeyError("VMFC isnt quite ready for primetime!")

            # ret.update({"vmfc" + str(nbody): [] for nbody in range(1, max_nbody)})
            # nbody_range = list(range(1, max_nbody))
            # for nbody in nbody_range:
            #     for cp_combos in it.combinations(fragment_range, nbody):
            #         basis_tuple = tuple(cp_combos)
            #         for interior_nbody in nbody_range:
            #             for x in it.combinations(cp_combos, interior_nbody):
            #                 ghost = list(set(basis_tuple) - set(x))
            #                 ret["vmfc" + str(interior_nbody)].append((mol.get_fragment(x, ghost), 1.0))

        # Add in the maximal position
        if do_default:
            ret["default"] = [(mol, 1.0)]

        if do_cp:
            ret["cp"] = [(mol, 1.0)]

        # if do_vmfc:
        #     ret["vmfc"] = [(mol, 1.0)]

        return ret



"""
Utilities and base functions for Services.
"""

import abc
import json
from typing import Any, Dict, List, Set, Tuple

from pydantic import BaseModel

from qcfractal.procedures import get_procedure_parser
from ..interface.models.rest_models import TaskQueuePOSTBody


class BaseService(BaseModel, abc.ABC):

    storage_socket: Any

    # Base information requiered by the class
    id: str = None
    hash_index: str
    status: str
    service: str
    program: str
    procedure: str

    @classmethod
    @abc.abstractmethod
    def initialize_from_api(cls, storage_socket, meta, molecule):
        """
        Initalizes a Service from the API
        """

    def dict(self, include=None, exclude=None, by_alias=False) -> Dict[str, Any]:
        return BaseModel.dict(self, exclude={"storage_socket"})

    def json_dict(self) -> str:
        return json.loads(self.json())

    @abc.abstractmethod
    def iterate(self):
        """
        Takes a "step" of the service. Should return False if not finished
        """


class TaskManager(BaseModel):

    required_tasks: Dict[str, str] = {}

    def done(self, storage_socket) -> bool:
        """
        Check if requested tasks are complete
        """

        if len(self.required_tasks) == 0:
            return True

        task_query = storage_socket.get_procedures_by_id(
            id=list(self.required_tasks.values()),
            projection={"status": True,
                        "error": True})

        if len(task_query["data"]) != len(self.required_tasks):
            return False

        elif "ERROR" in set(x["status"] for x in task_query["data"]):
            for x in task_query["data"]:
                if x["status"] != "ERROR":
                    continue
            raise KeyError("All tasks did not execute successfully.")

        return True

    def get_tasks(self, storage_socket) -> Dict[str, Any]:
        """
        Pulls currently held tasks
        """

        ret = {}
        for k, id in self.required_tasks.items():
            ret[k] = storage_socket.get_procedures_by_id(id=id)["data"][0]

        return ret

    def submit_tasks(self, storage_socket, procedure_type: str, tasks: Dict[str, Any]) -> bool:
        """
        Submits new tasks to the queue and provides a waiter until there are done.
        """
        procedure_parser = get_procedure_parser(procedure_type, storage_socket)

        required_tasks = {}

        # Add in all new tasks
        for key, packet in tasks.items():
            packet = TaskQueuePOSTBody(**packet)

            # Turn packet into a full task, if there are duplicates, get the ID
            r = procedure_parser.submit_tasks(packet)

            if len(r["meta"]["errors"]):
                raise KeyError("Problem submitting task: {}.".format(errors))

            required_tasks[key] = r["data"]["ids"][0]

        self.required_tasks = required_tasks

        return True


def expand_ndimensional_grid(dimensions: Tuple[int, ...], seeds: Set[Tuple[int, ...]],
                             complete: Set[Tuple[int, ...]]) -> List[Tuple[Tuple[int, ...], Tuple[int, ...]]]:
    """
    Expands an n-dimensional key/value grid

    Example:
    >>> expand_ndimensional_grid((3, 3), {(1, 1)}, set())
    [((1, 1), (0, 1)), ((1, 1), (2, 1)), ((1, 1), (1, 0)), ((1, 1), (1, 2))]
    """

    dimensions = tuple(dimensions)
    compute = set()
    connections = []

    for d in range(len(dimensions)):

        # Loop over all compute seeds
        for seed in seeds:

            # Iterate both directions
            for disp in [-1, 1]:
                new_dim = seed[d] + disp

                # Bound check
                if new_dim >= dimensions[d]:
                    continue
                if new_dim < 0:
                    continue

                new = list(seed)
                new[d] = new_dim
                new = tuple(new)

                # Push out duplicates from both new compute and copmlete
                if new in compute:
                    continue
                if new in complete:
                    continue

                compute |= {new}
                connections.append((seed, new))

    return connections



"""
This tests the Mongoenegine performance when handling Results and procedures

"""

import mongoengine as db
from qcfractal.storage_sockets.models import Molecule, Result, Keywords
from qcfractal.storage_sockets.mongoengine_socket import MongoengineSocket
import qcfractal.interface as portal
from time import time
import pymongo
from pymongo import MongoClient


db_name = 'bench_qc_mongoengine'
db_client = db.connect(db_name)
db_client.drop_database(db_name)

mongoengine_socket = MongoengineSocket("mongodb://localhost", db_name)
# mongoengine_socket.mongoengine_client.drop_database('bech_qc_mongoengine')

pymongo_client = MongoClient()[db_name]

n_mol = 1000
n_results = 1000
n_query = 1000


def insert_molecules(n_mol):
    # add Molecules
    water = portal.data.get_molecule("water_dimer_minima.psimol").to_json()

    mol_data = {}
    # Add Molecule using pymongo
    for i in range(n_mol):
        tmp = water.copy()
        tmp['molecular_charge'] = i
        mol_data['water'+str(i)] = tmp

    mongoengine_socket.add_molecules(mol_data)


def query_molecule(n_mol):
    for i in range(n_mol):
        Molecule.objects.first()

def insert_results(n_results, mol):

    # repeat searching for the molecule
    for i in range(n_results):
        # mol = Molecule.objects.first()  # one DB access
        # option = Keywords.objects().first()
        data = {
            "molecule": mol.id,
            "method": str(i),
            "basis": "B1",
            "keywords": None,
            "program": "P1",
            "driver": "energy",
            "other_data": 5,
        }
        results = Result(**data).save()  # second DB access
    return results


def bulk_insert_results(n_results, mol):

    results = []
    for i in range(n_results):
        # mol = Molecule.objects.first()  # one DB access
        # option = Keywords.objects().first()
        data = {
            "molecule": mol,
            "method": str(i),
            "basis": "Bulk",
            "keywords": None,
            "program": "P1",
            "driver": "energy",
            "other_data": 5,
        }
        results.append(Result(**data))
    return Result.objects.insert(results)


def bulk_insert_results_pymongo(n_results, mol):

    results = []
    for i in range(n_results):
        # mol = Molecule.objects.first()  # one DB access
        # option = Keywords.objects().first()
        data = {
            "molecule": mol.id,
            "method": str(i),
            "basis": "Bulk pymongo",
            "keywords": None,
            "program": "P1",
            "driver": "energy",
            "other_data": 5,
        }
        results.append(data)

    # results = Result._get_collection().insert_many(results, ordered=False)
    ret = pymongo_client.results.insert_many(results, ordered=False)

    return ret


def duplicate_results(n_results, mol):
    """Half the documents are duplicates"""

    tosave_results = []
    for i in range(n_results):
        # mol = Molecule.objects.first()  # one DB access
        # option = Keywords.objects().first()
        data = {
            "molecule": mol,
            "method": str(i + int(n_results/2)),
            "basis": "Bulk",
            "keywords": None,
            "program": "P1",
            "driver": "energy",
            "other_data": 5,
        }
        found = Result.objects(**data).first()
        if not found:
            tosave_results.append(Result(**data))

    Result.objects.insert(tosave_results)
    print('Duplciates: ', len(tosave_results))


def duplicate_results_pymongo(n_results, mol):
    """Half the documents are duplicates"""

    results = []
    for i in range(n_results):
        # mol = Molecule.objects.first()  # one DB access
        # option = Keywords.objects().first()
        data = {
            "molecule": mol.id,
            "method": str(i + int(n_results/2)),
            "basis": "Bulk pymongo",
            "keywords": None,
            "program": "P1",
            "driver": "energy",
            "other_data": 5,
        }
        results.append(data)

    try:
        # results = Result._get_collection().insert_many(results, ordered=False)
        pymongo_client.results.insert_many(results, ordered=False)
    except pymongo.errors.BulkWriteError as err:
        print('number of inserted: ', err.details["nInserted"])

    return results


def query_results(n_query, mol):
    for i in range(n_query):
        # mol = Molecule.objects.first()  # --> the overhead in this query
        # option = Keywords.objects(program='Psi4').first()  # or [0], throws ex
        query = {
            "molecule": mol,
            "method": str(i),
            "basis": "B1",
            "keywords": None,
            "program": "P1",
            "driver": "energy",
        }
        Result.objects(**query)  # second DB access


def query_results_pymongo(n_query, mol):
    for i in range(n_query):
        # mol = Molecule.objects.first()  # --> the overhead in this query
        # option = Keywords.objects(program='Psi4').first()  # or [0], throws ex
        query = {
            "molecule": mol.id,
            "method": str(i),
            "basis": "B1",
            "keywords": None,
            "program": "P1",
            "driver": "energy",
        }
        # Result._get_collection().find_one(query)
        # db.connection.get_connection()[db_name]['result'].find_one(query)
        pymongo_client.results.find(query)

def bench():

    option = Keywords(program='Psi4', name='default').save()

    tstart = time()
    insert_molecules(n_mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Molecules inserted in an avg {:0.3f} ms / doc'.format(n_mol, dtime/n_mol))

    tstart = time()
    query_molecule(n_mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Query Molecules in an avg {:0.3f} ms / doc'.format(n_mol, dtime/n_mol))

    print('---------------------------')

    mol = Molecule.objects.first()

    tstart = time()
    insert_results(n_results, mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Results inserted in an avg {:0.3f} ms / doc'.format(n_results, dtime/n_results))

    # ---> Bulk in much faster
    tstart = time()
    bulk_insert_results(n_results, mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Bulk Results inserted in an avg {:0.3f} ms / doc'.format(n_results, dtime/n_results))

     # ---> Bulk in much faster
    tstart = time()
    bulk_insert_results_pymongo(n_results, mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Bulk Results inserted PYMONGO in an avg {:0.3f} ms / doc'.format(n_results, dtime/n_results))


    print('--------------------------')

    # ---> Duplicates
    tstart = time()
    duplicate_results(n_results, mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Duplicate Results inserted in an avg {:0.3f} ms / doc'.format(n_results, dtime/n_results))

    # ---> Duplicates
    tstart = time()
    duplicate_results_pymongo(n_results, mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Duplicate Results PYMONGO inserted in an avg {:0.3f} ms / doc'.format(n_results, dtime/n_results))

    print('--------------------------')

    tstart = time()
    query_results(n_query, mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Results queries in an avg {:0.3f} ms / doc'.format(n_query, dtime/n_query))

    tstart = time()
    query_results_pymongo(n_query, mol)
    dtime = (time() - tstart) * 1000  # msec
    print('{} Results queries PYMONGO in an avg {:0.3f} ms / doc'.format(n_query, dtime/n_query))


if __name__ == "__main__":
    bench()




"""
Tests for QCFractals CLI
"""
import os
import time

import pytest

from qcfractal import testing

#def _run_tests()
_options = {"coverage": True, "dump_stdout": True}
_pwd = os.path.dirname(os.path.abspath(__file__))


@testing.mark_slow
def test_cli_server_boot():
    port = "--port=" + str(testing.find_open_port())
    assert testing.run_process(["qcfractal-server", "mydb", port], interupt_after=10, **_options)


@testing.mark_slow
@testing.using_fireworks
def test_cli_server_local_boot():
    port = "--port=" + str(testing.find_open_port())
    args = ["qcfractal-server", "mydb", "--local-manager", port]
    assert testing.run_process(args, interupt_after=10, **_options)


@pytest.fixture(scope="module")
def active_server(request):
    port = str(testing.find_open_port())
    args = ["qcfractal-server", "mydb", "--port=" + port]
    with testing.popen(args, **_options) as server:
        time.sleep(2)

        server.test_uri_cli = "--fractal-uri=localhost:" + port
        yield server


@testing.mark_slow
def test_manager_local_testing_process():
    assert testing.run_process(["qcfractal-manager", "--test", "executor"], **_options)


@testing.mark_slow
def test_manager_executor_manager_boot(active_server):
    args = ["qcfractal-manager", active_server.test_uri_cli, "executor", "--nprocs=1"]
    assert testing.run_process(args, interupt_after=7, **_options)


@testing.mark_slow
@testing.using_dask
def test_manager_dask_manager_local_boot(active_server):
    args = ["qcfractal-manager", active_server.test_uri_cli, "dask", "--local-cluster", "--local-workers=1"]
    assert testing.run_process(args, interupt_after=7, **_options)


@testing.mark_slow
@testing.using_fireworks
def test_manager_fireworks_boot(active_server):
    args = ["qcfractal-manager", active_server.test_uri_cli, "fireworks"]
    assert testing.run_process(args, interupt_after=5, **_options)


@testing.mark_slow
@testing.using_fireworks
def test_manager_fireworks_config_boot(active_server):
    config_path = os.path.join(_pwd, "fw_config_boot.yaml")
    args = [
        "qcfractal-manager", active_server.test_uri_cli, "--rapidfire", "--config-file=" + config_path, "fireworks"
    ]
    assert testing.run_process(args, **_options)



"""
All procedures tasks involved in on-node computation.
"""

import json
from typing import Union

from . import procedures_util


class SingleResultTasks:
    """Single is a simple Result
     Unique by: driver, method, basis, option (the name in the options table),
     and program.
    """

    def __init__(self, storage):
        self.storage = storage

    def parse_input(self, data):
        """Parse input json into internally appropriate format


        Format of the input data:
        data = {
            "meta": {
                "procedure": "single",
                "driver": "energy",
                "method": "HF",
                "basis": "sto-3g",
                "keywords": "default",
                "program": "psi4"
                },
            },
            "data": ["mol_id_1", "mol_id_2", ...],
        }

        """

        # format the data
        inputs, errors = procedures_util.unpack_single_run_meta(self.storage, data.meta, data.data)

        # Insert new results stubs
        result_stub = json.dumps({k: data.meta[k] for k in ["driver", "method", "basis", "keywords", "program"]})

        # Grab the tag if available
        tag = data.meta.pop("tag", None)

        # Construct full tasks
        new_tasks = []
        results_ids = []
        existing_ids = []
        for inp in inputs:
            if inp is None:
                results_ids.append(None)
                continue

            # Build stub
            result_obj = json.loads(result_stub)
            result_obj["molecule"] = str(inp.molecule.id)
            ret = self.storage.add_results([result_obj])

            base_id = ret["data"][0]
            results_ids.append(base_id)

            # Task is complete
            if len(ret["meta"]["duplicates"]):
                existing_ids.append(base_id)
                continue

            # Build task object
            task = {
                "spec": {
                    "function": "qcengine.compute",  # todo: add defaults in models
                    "args": [json.loads(inp.json()), data.meta["program"]], # todo: json_dict should come from results
                    "kwargs": {}  # todo: add defaults in models
                },
                "hooks": [],  # todo: add defaults in models
                "tag": tag,
                "parser": "single",
                "base_result": ("results", base_id)
            }

            new_tasks.append(task)

        return new_tasks, results_ids, existing_ids, errors

    def submit_tasks(self, data):

        new_tasks, results_ids, existing_ids, errors = self.parse_input(data)

        self.storage.queue_submit(new_tasks)

        n_inserted = 0
        missing = []
        for num, x in enumerate(results_ids):
            if x is None:
                missing.append(num)
            else:
                n_inserted += 1

        results = {
            "meta": {
                "n_inserted": n_inserted,
                "duplicates": [],
                "validation_errors": [],
                "success": True,
                "error_description": False,
                "errors": errors
            },
            "data": {
                "ids": results_ids,
                "submitted": [x["base_result"][1] for x in new_tasks],
                "existing": existing_ids,
            }
        }

        return results

    def parse_output(self, data):

        # Add new runs to database
        # Parse out hooks and data to same key/value
        rdata = {}
        rhooks = {}
        for data, hooks in data:
            key = data["task_id"]
            rdata[key] = data
            if len(hooks):
                rhooks[key] = hooks

        # Add results to database
        results = procedures_util.parse_single_runs(self.storage, rdata)

        ret = self.storage.add_results(list(results.values()), update_existing=True)

        # Sort out hook data
        hook_data = procedures_util.parse_hooks(results, rhooks)

        # Create a list of (queue_id, located) to update the queue with
        completed = list(results.keys())

        errors = []
        if len(ret["meta"]["errors"]):
            # errors = [(k, "Duplicate results found")]
            raise ValueError("TODO: Cannot yet handle queue result duplicates.")

        return completed, errors, hook_data


# ----------------------------------------------------------------------------


class OptimizationTasks(SingleResultTasks):
    """
    Optimization task manipulation
    """

    def parse_input(self, data, duplicate_id="hash_index"):
        """

        json_data = {
            "meta": {
                "procedure": "optimization",
                "option": "default",
                "program": "geometric",
                "qc_meta": {
                    "driver": "energy",
                    "method": "HF",
                    "basis": "sto-3g",
                    "keywords": "default",
                    "program": "psi4"
                },
            },
            "data": ["mol_id_1", "mol_id_2", ...],
        }

        qc_schema_input = {
            "molecule": {
                "geometry": [
                    0.0,  0.0, -0.6,
                    0.0,  0.0,  0.6,
                ],
                "symbols": ["H", "H"],
                "connectivity": [[0, 1, 1]]
            },
            "driver": "gradient",
            "model": {
                "method": "HF",
                "basis": "sto-3g"
            },
            "keywords": {},
        }
        json_data = {
            "keywords": {
                "coordsys": "tric",
                "maxiter": 100,
                "program": "psi4"
            },
        }

        """

        # Unpack individual QC tasks
        inputs, errors = procedures_util.unpack_single_run_meta(self.storage, data.meta["qc_meta"], data.data)

        # Unpack options
        if data.meta["keywords"] is None:
            keyword_set = {}
            keyword_id = None
        else:
            keyword_set = self.storage.get_add_keywords_mixed([data.meta["keywords"]])["data"][0]
            keyword_id = keyword_set["id"]
            keyword_set = keyword_set["values"]

        keyword_set["program"] = data.meta["qc_meta"]["program"]
        template = json.dumps({"keywords": keyword_set, "qcfractal_tags": data.meta})

        tag = data.meta.pop("tag", None)

        new_tasks = []
        results_ids = []
        existing_ids = []
        for inp in inputs:
            if inp is None:
                results_ids.append(None)
                continue

            inp = json.loads(inp.json())

            # Coerce qc_template information
            packet = json.loads(template)
            packet["initial_molecule"] = inp.pop("molecule")
            packet["input_specification"] = inp
            packet["procedure"] = "optimization"
            packet["program"] = data.meta["program"]

            single_keys = data.meta["qc_meta"].copy()
            single_keys["molecule"] = packet["initial_molecule"]["id"]

            # Add to args document to carry through to self.storage
            hash_index = procedures_util.hash_procedure_keys({
                "type": "optimization",
                "program": data.meta["program"],
                "keywords": keyword_id,
                "single_key": single_keys,
            })
            packet["hash_index"] = hash_index

            ret = self.storage.add_procedures([packet])
            base_id = ret["data"][0]
            results_ids.append(base_id)

            # Task is complete
            if len(ret["meta"]["duplicates"]):
                existing_ids.append(base_id)
                continue

            # Build task object
            task = {
                "spec": {
                    "function": "qcengine.compute_procedure",
                    "args": [packet, data.meta["program"]],
                    "kwargs": {}
                },
                "hooks": [],
                "tag": tag,
                "parser": "optimization",
                "base_result": ("procedure", base_id)
            }

            new_tasks.append(task)

        return new_tasks, results_ids, existing_ids, errors


    def parse_output(self, data):
        """Save the results of the procedure.
        It must make sure to save the results in the results table
        including the task_id in the TaskQueue table
        """

        new_procedures = {}
        new_hooks = {}

        # Each optimization is a unique entry:
        for procedure, hooks in data:
            task_id = procedure["task_id"]

            # Convert start/stop molecules to hash
            mols = {"initial": procedure["initial_molecule"], "final": procedure["final_molecule"]}
            mol_keys = self.storage.add_molecules(mols)["data"]
            procedure["initial_molecule"] = mol_keys["initial"]
            procedure["final_molecule"] = mol_keys["final"]

            # Parse trajectory computations and add task_id
            traj_dict = {k: v for k, v in enumerate(procedure["trajectory"])}
            results = procedures_util.parse_single_runs(self.storage, traj_dict)
            for k, v in results.items():
                v["task_id"] = task_id

            # Add trajectory results and return ids
            ret = self.storage.add_results(list(results.values()))
            procedure["trajectory"] = ret["data"]

            # Coerce tags
            procedure.update(procedure["qcfractal_tags"])
            del procedure["input_specification"]
            del procedure["qcfractal_tags"]
            # print("Adding optimization result")
            # print(json.dumps(v, indent=2))
            new_procedures[task_id] = procedure
            if len(hooks):
                new_hooks[task_id] = hooks

            self.storage.update_procedure(procedure["hash_index"], procedure)

        # Create a list of (queue_id, located) to update the queue with
        completed = list(new_procedures.keys())

        errors = []
        # if len(ret["meta"]["errors"]):
        #     # errors = [(k, "Duplicate results found")]
        #     raise ValueError("TODO: Cannot yet handle queue result duplicates.")

        hook_data = procedures_util.parse_hooks(new_procedures, new_hooks)

        # return (ret, hook_data)
        return completed, errors, hook_data


# ----------------------------------------------------------------------------

supported_procedures = Union[SingleResultTasks, OptimizationTasks]


def get_procedure_parser(procedure_type: str, storage) -> supported_procedures:
    """A factory methods that returns the approperiate parser class
    for the supported procedure types (like single and optimization)

    Parameters
    ---------
    procedure_type: str, 'single' or 'optimization'
    storage: storage socket object
        such as MongoengineSocket object

    Returns
    -------
    A parser class corresponding to the procedure_type:
        'single' --> SingleResultTasks
        'optimization' --> OptimizationTasks
    """

    if procedure_type == 'single':
        return SingleResultTasks(storage)
    elif procedure_type == 'optimization':
        return OptimizationTasks(storage)
    else:
        raise KeyError("Procedure type ({}) is not suported yet.".format(procedure_type))



"""Provides an interface the QCDB Server instance"""

import json
import os
from collections import defaultdict
from typing import List, Union, Dict, Any, Optional

import requests

from . import orm
from .collections import collection_factory

from .models.common_models import Molecule
from .models.rest_models import (
    MoleculeGETBody, MoleculeGETResponse, MoleculePOSTBody, MoleculePOSTResponse,
    KeywordGETBody, KeywordGETResponse, KeywordPOSTBody, KeywordPOSTResponse,
    CollectionGETBody, CollectionGETResponse, CollectionPOSTBody,CollectionPOSTResponse,
    ResultGETBody, ResultGETResponse, ProcedureGETBody, ProcedureGETReponse,
    TaskQueueGETBody, TaskQueueGETResponse, TaskQueuePOSTBody, TaskQueuePOSTResponse,
    ServiceQueuePOSTBody, ServiceQueuePOSTResponse, ServiceQueueGETBody, ServiceQueueGETResponse) # yapf: disable
from .models.gridoptimization import GridOptimizationInput
from .models.torsiondrive import TorsionDriveInput


class FractalClient(object):
    def __init__(self, address: Any, username: Optional[str]=None, password: Optional[str]=None, verify: bool=True):
        """Initializes a FractalClient instance from an address and verification information.

        Parameters
        ----------
        address : str or FractalServer
            The IP and port of the FractalServer instance ("192.168.1.1:8888") or
            a FractalServer instance
        username : None, optional
            The username to authenticate with.
        password : None, optional
            The password to authenticate with.
        verify : bool, optional
            Verifies the SSL connection with a third party server. This may be False if a
            FractalServer was not provided a SSL certificate and defaults back to self-signed
            SSL keys.
        """

        if hasattr(address, "get_address"):
            # We are a FractalServer-like object
            verify = address.client_verify
            address = address.get_address()

        if "http" not in address:
            address = "https://" + address

        # If we are `http`, ignore all SSL directives
        if not address.startswith("https"):
            self._verify = True

        if not address.endswith("/"):
            address += "/"

        self.address = address
        self.username = username
        self._verify = verify
        self._headers = {}

        # If no 3rd party verification, quiet urllib
        if self._verify is False:
            from urllib3.exceptions import InsecureRequestWarning
            requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

        if (username is not None) or (password is not None):
            self._headers["Authorization"] = json.dumps({"username": username, "password": password})

        self._headers["content_type"] = 'application/json'

        # Try to connect and pull general data
        self.server_info = self._request("get", "information", {}).json()

        self.server_name = self.server_info["name"]

    def __str__(self) -> str:
        """A short short representation of the current FractalClient.

        Returns
        -------
        str
            The desired representation.
        """
        ret = "FractalClient(server_name='{}', address='{}', username='{}')".format(
            self.server_name, self.address, self.username)
        return ret

    def _request(self, method: str, service: str, payload: Dict[str, Any]=None, *, data: str=None,
                 noraise: bool=False):

        addr = self.address + service
        try:
            if method == "get":
                r = requests.get(addr, json=payload, data=data, headers=self._headers, verify=self._verify)
            elif method == "post":
                r = requests.post(addr, json=payload, data=data, headers=self._headers, verify=self._verify)
            elif method == "put":
                r = requests.put(addr, json=payload, data=data, headers=self._headers, verify=self._verify)
            else:
                raise KeyError("Method not understood: '{}'".format(method))
        except requests.exceptions.SSLError as exc:
            error_msg = (
                "\n\nSSL handshake failed. This is likely caused by a failure to retrive 3rd party SSL certificates.\n"
                "If you trust the server you are connecting to, try 'FractalClient(... verify=False)'")
            raise requests.exceptions.SSLError(error_msg)
        except requests.exceptions.ConnectionError as exc:
            error_msg = (
                "\n\nCould not connect to server {}, please check the address and try again.".format(self.address))
            raise requests.exceptions.ConnectionError(error_msg)

        if (r.status_code != 200) and (not noraise):
            raise requests.exceptions.HTTPError("Server communication failure. Reason: {}".format(r.reason))

        return r

    @classmethod
    def from_file(cls, load_path: Optional[str]=None):
        """Creates a new FractalClient from file. If no path is passed in searches
        current working directory and ~.qca/ for "qcportal_config.yaml"

        Parameters
        ----------
        load_path : str, dict, optional
            Path to find "qcportal_config.yaml", the filename, or a dictionary containing keys
            ["address", "username", "password", "verify"]

        """

        # Search canonical paths
        if load_path is None:
            test_paths = [os.getcwd(), os.path.join(os.path.expanduser('~'), ".qca")]

            for path in test_paths:
                local_path = os.path.join(path, "qcportal_config.yaml")
                if os.path.exists(local_path):
                    load_path = local_path
                    break

            if load_path is None:
                raise FileNotFoundError("Could not find `qcportal_config.yaml` in the following paths:\n    {}".format(
                    ", ".join(test_paths)))

        # Load if string, or use if dict
        if isinstance(load_path, str):
            load_path = os.path.join(os.path.expanduser(load_path))

            # Gave folder, not file
            if os.path.isdir(load_path):
                load_path = os.path.join(load_path, "qcportal_config.yaml")

            with open(load_path, "r") as handle:
                import yaml
                data = yaml.load(handle)

        elif isinstance(load_path, dict):
            data = load_path
        else:
            raise TypeError("Could not infer data from load_path of type {}".format(type(load_path)))

        if "address" not in data:
            raise KeyError("Config file must at least contain a address field.")

        address = data["address"]
        username = data.get("username", None)
        password = data.get("password", None)
        verify = data.get("verify", True)

        return cls(address, username=username, password=password, verify=verify)

    def server_information(self) -> Dict[str, str]:
        return json.loads(json.dumps(self.server_info))

    ### Molecule section

    def get_molecules(self, mol_list: List[str], index: str="id", full_return: bool=False) -> Dict[str, Any]:
        """Get molecules from the Server.

        Parameters
        ----------
        mol_list : list of str
            Either molecule Id's or molecule hashes to query.
        index : str, ("id", "hash")
            The index to search on
        full_return : bool, optional
            Flags to return all metadata or only the query.

        Returns
        -------
        list of molecule JSON
            Returns all found molecules.
        """

        if isinstance(mol_list, str):
            mol_list = [mol_list]

        body = MoleculeGETBody(data=mol_list, meta={"index": index.lower()})
        r = self._request("get", "molecule", data=body.json())
        r = MoleculeGETResponse.parse_raw(r.text)

        if full_return:
            return r
        else:
            return r.data

    def add_molecules(self, mol_list: Dict[str, Molecule],
                      full_return: bool=False) -> Union[List[str], Dict[str, Any]]:
        """Adds molecules to the Server

        Parameters
        ----------
        mol_list : dict
            A (key: molecule) dictionary for the molecules to be added. The molecules can either be a
            Molecule class or a JSON Molecule representation.
        full_return : bool, optional
            Flags to return all metadata or only the submitted ids.

        Returns
        -------
        dict
            A (key: molecule id) dictionary of added molecules.

        """

        body = MoleculePOSTBody(meta={}, data=mol_list)
        r = self._request("post", "molecule", data=body.json())
        r = MoleculePOSTResponse.parse_raw(r.text)

        if full_return:
            return r
        else:
            return r.data

    ### Keywords section

    def get_keywords(self, opt_list):

        body = KeywordGETBody(meta={}, data=opt_list)
        r = self._request("get", "keyword", data=body.json())
        r = KeywordGETResponse.parse_raw(r.text)

        return r.data

    def add_keywords(self, opt_list: List[Dict[str, Any]],
                     full_return: bool=False) -> Union[List[str], Dict[str, Any]]:

        body = KeywordPOSTBody(meta={}, data=opt_list)
        r = self._request("post", "keyword", data=body.json())
        r = KeywordPOSTResponse.parse_raw(r.text)

        if full_return:
            return r
        else:
            return r.data

    ### Collections section

    def list_collections(self, collection_type: Optional[str]=None) -> Dict[str, Any]:
        """Lists the available collections currently on the server.

        Parameters
        ----------
        collection_type : None, optional
            If `None` all collection types will be returned, otherwise only the
            specified collection type will be returned

        Returns
        -------
        dict
            A dictionary containing the available collection types.
        """

        query = {}
        if collection_type is not None:
            query = {"collection": collection_type.lower()}

        payload = {"meta": {"projection": {"name": True, "collection": True}}, "data": query}
        r = self._request("get", "collection", payload)

        if collection_type is None:
            ret = defaultdict(list)
            for entry in r.json()["data"]:
                ret[entry["collection"]].append(entry["name"])
            return dict(ret)
        else:
            return [x["name"] for x in r.json()["data"]]

    def get_collection(self, collection_type: str, name: str, full_return: bool=False):
        """Acquires a given collection from the server

        Parameters
        ----------
        collection_type : str
            The collection type to be accessed
        name : str
            The name of the collection to be accessed
        full_return : bool, optional
            If False, returns a Collection object otherwise returns raw JSON

        Returns
        -------
        Collection
            A Collection object if the given collection was found otherwise returns `None`.
        """

        body = CollectionGETBody(meta={}, data={"collection": collection_type, "name": name})
        r = self._request("get", "collection", data=body.json())
        cols = CollectionGETResponse.parse_raw(r.text)
        if full_return:
            return cols
        else:
            # If nothing found
            if len(cols.data):
                return collection_factory(cols.data[0], client=self)
            else:
                raise KeyError("Collection '{}:{}' not found.".format(collection_type, name))

    def add_collection(self, collection: Dict[str, Any], overwrite: bool=False, full_return: bool=False):

        # Can take in either molecule or lists

        if overwrite and ("id" not in collection or collection['id'] == 'local'):
            raise KeyError("Attempting to overwrite collection, but no server ID found (cannot use 'local').")

        payload = {"meta": {"overwrite": overwrite}, "data": collection}
        body = CollectionPOSTBody(**payload)
        r = self._request("post", "collection", data=body.json())
        assert r.status_code == 200
        r = CollectionPOSTResponse.parse_raw(r.text)

        if full_return:
            return r
        else:
            return r.data

    ### Results section

    def get_results(self, **kwargs):
        projection = kwargs.pop("projection", None)
        return_full = kwargs.pop("return_full", False)

        payload = {"meta": {}, "data": kwargs}
        if projection is not None:
            payload["meta"]["projection"] = projection

        body = ResultGETBody(**payload)
        r = self._request("get", "result", data=body.json())
        r = ResultGETResponse.parse_raw(r.text)

        if return_full:
            return r
        else:
            return r.data

    def check_results(self, **kwargs):

        kwargs["status"] = None
        if "projection" in kwargs:
            kwargs["projection"]["status"] = True
        else:
            kwargs["projection"] = {"status": True}
        return self.get_results(**kwargs)

    def get_procedures(self, procedure_query: Dict[str, Any], return_objects: bool=True):

        body = ProcedureGETBody(data=procedure_query)
        r = self._request("get", "procedure", data=body.json())
        r = ProcedureGETReponse.parse_raw(r.text)

        if return_objects:
            ret = []
            for packet in r.data:
                tmp = orm.build_orm(packet, client=self)
                ret.append(tmp)
            return ret
        else:
            # Equivalent to full_return from other gets
            return r

    # Must compute results?
    # def add_results(self, db, full_return=False):

    #     # Can take in either molecule or lists

    #     payload = {"meta": {}, "data": {}}
    #     payload["data"] = db

    #     r = requests.post(self._result_addr, json=payload)
    #     assert r.status_code == 200

    #     if full_return:
    #         return r.json()
    #     else:
    #         return r.json()["data"]

    ### Compute section

    def add_compute(self,
                    program: str,
                    method: str,
                    basis: str,
                    driver: str,
                    keywords: Union[str, None],
                    molecule_id: Union[str, Molecule, List[Union[str, Molecule]]],
                    return_full: bool=False,
                    tag: str=None) -> Union[TaskQueuePOSTResponse, TaskQueuePOSTResponse.Data]:

        # Always a list
        if not isinstance(molecule_id, list):
            molecule_id = [molecule_id]

        payload = {
            "meta": {
                "procedure": "single",
                "driver": driver,
                "program": program,
                "method": method,
                "basis": basis,
                "keywords": keywords,
                "tag": tag,
            },
            "data": molecule_id
        }

        body = TaskQueuePOSTBody(**payload)

        r = self._request("post", "task_queue", data=body.json())
        r = TaskQueuePOSTResponse.parse_raw(r.text)

        if return_full:
            return r
        else:
            return r.data

    def add_procedure(self,
                      procedure: str,
                      program: str,
                      program_options: Dict[str, Any],
                      molecule_id: List[str],
                      return_full: bool=False):

        # Always a list
        if isinstance(molecule_id, str):
            molecule_id = [molecule_id]

        payload = {
            "meta": {
                "procedure": procedure,
                "program": program,
            },
            "data": molecule_id
        }
        payload["meta"].update(program_options)

        r = self._request("post", "task_queue", payload)
        r = TaskQueuePOSTResponse.parse_raw(r.text)

        if return_full:
            return r
        else:
            return r.data

    def check_tasks(self, query: Dict[str, Any], projection: Optional[Dict[str, Any]]=None, return_full: bool=False):
        """Checks the status of tasks in the Fractal queue.

        Parameters
        ----------
        query : dict
            A query to find tasks
        projection: dict, optional
            Projection of data to call from the database
        return_full : bool, optional
            Returns the full JSON return if True

        Returns
        -------
        list of dict
            A dictionary of each match that contains the current status
            and, if an error has occured, the error message.

        >>> client.check_tasks({"id": "5bd35af47b878715165f8225"})
        [{"status": "WAITING"}]
        """

        payload = {"meta": {"projection": projection}, "data": query}
        body = TaskQueueGETBody(**payload)

        r = self._request("get", "task_queue", data=body.json())
        r = TaskQueueGETResponse.parse_raw(r.text)

        if return_full:
            return r
        else:
            return r.data

    def add_service(self, service: Union[GridOptimizationInput, TorsionDriveInput], return_full: bool=False):

        body = ServiceQueuePOSTBody(meta={}, data=service)

        r = self._request("post", "service_queue", data=body.json())
        r = ServiceQueuePOSTResponse.parse_raw(r.text)

        if return_full:
            return r
        else:
            return r.data

    def check_services(self, query: Dict[str, Any], return_full: bool=False):
        """Checks the status of services in the Fractal queue.

        Parameters
        ----------
        query : dict
            A query to find services
        return_full : bool, optional
            Returns the full JSON return if True

        Returns
        -------
        list of dict
            A dictionary of each match that contains the current status
            and, if an error has occurred, the error message.

        >>> client.check_services({"id": "5bd35af47b878715165f8225"})
        [{"status": "RUNNING"}]
        """

        payload = {"meta": {}, "data": query}

        body = ServiceQueueGETBody(**payload)

        r = self._request("get", "service_queue", data=body.json())
        r = ServiceQueueGETResponse.parse_raw(r.text)

        if return_full:
            return r
        else:
            return r.data



"""
Either pull in QCEl models or local models
"""


from . import common_models
from . import gridoptimization
from . import rest_models
from . import torsiondrive
from .common_models import Molecule, KeywordSet



"""
All json-schema definitions for the project
"""

import copy

__all__ = ["get_definition"]

_definitions = {
    "provenance": {
        "properties": {
            "creator": {
                "description": "The name of the person or program who created this object.",
                "type": "string"
            },
            "version": {
                "description": "The version of the program which created this object, blank otherwise.",
                "type": "string"
            },
            "routine": {
                "description": "The routine of the program which created this object, blank otherwise.",
                "type": "string"
            }
        },
        "required": ["creator"],
        "description": "A short provenance of the object.",
        "additionalProperties": True
    }
}


def get_definition(definition):
    if definition not in _definitions:
        raise KeyError("Definition '{}' not present.".format(definition))

    return copy.deepcopy(_definitions[definition])



"""
Tests the server compute capabilities.
"""

import pytest
import requests

import qcfractal.interface as portal
from qcfractal import testing
from qcfractal.testing import fractal_compute_server, reset_server_database, using_psi4, using_rdkit


@pytest.mark.parametrize("data", [
    pytest.param(("psi4", "HF", "sto-3g"), id="psi4", marks=using_psi4),
    pytest.param(("rdkit", "UFF", None), id="rdkit", marks=using_rdkit)
])
def test_task_molecule_no_orientation(data, fractal_compute_server):
    """
    Molecule orientation should not change on compute
    """

    # Reset database each run
    reset_server_database(fractal_compute_server)

    client = portal.FractalClient(fractal_compute_server)

    mol = portal.Molecule(symbols=["H", "H"], geometry=[0, 0, 0, 0, 5, 0], connectivity=[(0, 1, 1)])

    mol_id = client.add_molecules({"h2": mol})["h2"]

    program, method, basis = data
    ret = client.add_compute(program, method, basis, "energy", None, [mol_id])

    # Manually handle the compute
    fractal_compute_server.await_results()

    # Check for the single result
    ret = client.get_results(id=ret.submitted)
    assert len(ret) == 1
    assert ret[0]["status"] == "COMPLETE"
    assert ret[0]["molecule"] == mol_id

    # Make sure no other molecule was added
    ret = client.get_molecules(["H2"], index="molecular_formula")
    assert len(ret) == 1
    assert ret[0].id == mol_id


@testing.using_rdkit
def test_task_error(fractal_compute_server):
    client = portal.FractalClient(fractal_compute_server)

    mol = portal.models.common_models.Molecule(**{
        "geometry": [0, 0, 0],
        "symbols": ["He"]
    })
    # Cookiemonster is an invalid method
    ret = client.add_compute("rdkit", "cookiemonster", "", "energy", None, [mol])

    # Manually handle the compute
    fractal_compute_server.await_results()

    # Check for error
    ret = client.check_results(id=ret.submitted)

    assert len(ret) == 1
    assert ret[0]["status"] == "ERROR"
    # assert "run_rdkit" in ret[0]["error"]["error_message"]
